import argparse
import logging
from typing import Any, Dict, List, Optional

import annofabapi
from annofabapi.models import OrganizationMember, ProjectMember, ProjectMemberRole

import annofabcli
from annofabcli import AnnofabApiFacade
from annofabcli.common.cli import AbstractCommandLineInterface, build_annofabapi_resource_and_login

logger = logging.getLogger(__name__)


class CopyProjectMembers(AbstractCommandLineInterface):
    """
    プロジェクトメンバをコピーする
    """

    def __init__(self, service: annofabapi.Resource, facade: AnnofabApiFacade, args: argparse.Namespace):
        super().__init__(service, facade, args)

        src_project_id = args.src_project_id
        dest_project_id = args.dest_project_id
        project_title1 = self.facade.get_project_title(src_project_id)
        project_title2 = self.facade.get_project_title(dest_project_id)

        self.src_project_title = project_title1
        self.dest_project_title = project_title2

    def validate_projects(self, src_project_id: str, dest_project_id: str):
        """
        適切なRoleが付与されているかを確認する。

        Raises:
             AuthorizationError: 自分自身のRoleがいずれかのRoleにも合致しなければ、AuthorizationErrorが発生する。

        """

        super().validate_project(src_project_id, project_member_roles=None)
        super().validate_project(dest_project_id, project_member_roles=[ProjectMemberRole.OWNER])

    def get_organization_members_from_project_id(self, project_id: str) -> List[OrganizationMember]:
        organization_name = self.facade.get_organization_name_from_project_id(project_id)
        return self.service.wrapper.get_all_organization_members(organization_name)

    @staticmethod
    def find_member(members: List[Dict[str, Any]], account_id: str) -> Optional[Dict[str, Any]]:
        for m in members:
            if m["account_id"] == account_id:
                return m

        return None

    def copy_project_members(self, src_project_id: str, dest_project_id: str, delete_dest: bool = False):
        """
        プロジェクトメンバを、別のプロジェクトにコピーする。

        Args:
            src_project_id: コピー元のproject_id
            dest_project_id: コピー先のproject_id
            delete_dest: Trueならばコピー先にしか存在しないプロジェクトメンバを削除する。

        """

        self.validate_projects(src_project_id, dest_project_id)

        src_project_members = self.service.wrapper.get_all_project_members(src_project_id)
        src_organization_members = self.get_organization_members_from_project_id(src_project_id)
        # 組織外のメンバが含まれている可能性があるので、除去する
        src_project_members = [
            e for e in src_project_members if self.find_member(src_organization_members, e["account_id"]) is not None
        ]

        dest_project_members = self.service.wrapper.get_all_project_members(dest_project_id)
        dest_organization_members = self.get_organization_members_from_project_id(dest_project_id)
        dest_organization_name = self.facade.get_organization_name_from_project_id(dest_project_id)

        # 追加対象のプロジェクトメンバ
        added_members = []
        for member in src_project_members:
            account_id = member["account_id"]
            if self.find_member(dest_organization_members, account_id) is None:
                # コピー先の組織メンバでないので、コピーしない
                logger.debug(
                    f"コピーしないメンバ: {member['user_id']} , {member['username']} : "
                    f"(コピー先の所属組織 {dest_organization_name} の組織メンバでないため)"
                )
                continue

            added_members.append(member)

        for member in added_members:
            logger.debug(f"{self.dest_project_title} プロジェクトに追加/更新するメンバ: {member['user_id']} , {member['username']}")

        # 更新対象のプロジェクトメンバ
        updated_members = added_members
        deleted_dest_members: List[ProjectMember] = []

        if delete_dest:
            # コピー先にしかいないメンバを削除する
            src_account_ids = [e["account_id"] for e in src_project_members]
            deleted_dest_members = [e for e in dest_project_members if e["account_id"] not in src_account_ids]

            def to_inactive(arg_member):
                arg_member["member_status"] = "inactive"
                return arg_member

            deleted_dest_members = list(map(to_inactive, deleted_dest_members))
            for member in deleted_dest_members:
                logger.debug(f"{self.dest_project_title} プロジェクトから削除するメンバ: {member['user_id']} , {member['username']}")

            updated_members = updated_members + deleted_dest_members

        if len(added_members) > 0:
            if self.confirm_processing(
                f"{self.src_project_title} プロジェクトのメンバを、"
                f"{self.dest_project_title} にコピーしますか？"
                f"追加対象: {len(added_members)} 件, 削除対象: {len(deleted_dest_members)} 件"
            ):
                self.service.wrapper.put_project_members(dest_project_id, updated_members)
        else:
            logger.info(f"{self.dest_project_title}のプロジェクトメンバに追加/更新はありません。")

    def main(self):
        args = self.args
        self.copy_project_members(args.src_project_id, args.dest_project_id, delete_dest=args.delete_dest)


def main(args):
    service = build_annofabapi_resource_and_login(args)
    facade = AnnofabApiFacade(service)
    CopyProjectMembers(service, facade, args).main()


def parse_args(parser: argparse.ArgumentParser):
    parser.add_argument("src_project_id", type=str, help="コピー元のプロジェクトのproject_id")
    parser.add_argument("dest_project_id", type=str, help="コピー先のプロジェクトのproject_id")

    parser.add_argument("--delete_dest", action="store_true", help="コピー先のプロジェクトにしか存在しないプロジェクトメンバを削除します。")

    parser.set_defaults(subcommand_func=main)


def add_parser(subparsers: argparse._SubParsersAction):
    subcommand_name = "copy"
    subcommand_help = "プロジェクトメンバをコピーする。"
    description = "プロジェクトメンバをコピーする。"
    epilog = "コピー先のプロジェクトに対して、オーナロールを持つユーザで実行してください。"

    parser = annofabcli.common.cli.add_parser(subparsers, subcommand_name, subcommand_help, description, epilog=epilog)
    parse_args(parser)



"""
annofabapi.exceptions

This module contains the set of annofabapi exceptions.
"""

from typing import List, Optional

from annofabapi.models import OrganizationMemberRole, ProjectMemberRole


class AnnofabCliException(Exception):
    """
    annofabcliに関するException
    """


class AuthenticationError(AnnofabCliException):
    """
    AnnoFabの認証エラー
    """

    def __init__(self, loing_user_id: str):
        msg = f"AnnoFabにログインできませんでした。User ID: {loing_user_id}"
        super().__init__(msg)


class UpdatedFileForDownloadingError(AnnofabCliException):
    """
    ダウンロード対象ファイルの更新処理のエラー
    """

    def __init__(self, msg: Optional[str] = None):
        super().__init__(msg)


class DownloadingFileNotFoundError(AnnofabCliException):
    """
    ダウンロード対象のファイルが存在しないときのエラー
    """

    def __init__(self, msg: Optional[str] = None):
        super().__init__(msg)


class AuthorizationError(AnnofabCliException):
    pass


class ProjectAuthorizationError(AuthorizationError):
    """
    AnnoFabプロジェクトに関する認可エラー
    """

    def __init__(self, project_title: str, roles: List[ProjectMemberRole]):
        role_values = [e.value for e in roles]
        msg = f"プロジェクト: {project_title} に、ロール: {role_values} のいずれかが付与されていません。"
        super().__init__(msg)


class OrganizationAuthorizationError(AuthorizationError):
    """
    AnnoFab組織に関する認可エラー
    """

    def __init__(self, organization_name: str, roles: List[OrganizationMemberRole]):
        role_values = [e.value for e in roles]
        msg = f"組織: {organization_name} に、ロール: {role_values} のいずれかが付与されていません。"
        super().__init__(msg)



import argparse
import logging.handlers
import time
import uuid
from pathlib import Path

import pyquery

import annofabcli
from annofabcli import AnnofabApiFacade
from annofabcli.common.cli import AbstractCommandLineInterface, ArgumentParser, build_annofabapi_resource_and_login

logger = logging.getLogger(__name__)


class UploadInstruction(AbstractCommandLineInterface):
    """
    作業ガイドをアップロードする
    """

    def upload_html_to_instruction(self, project_id: str, html_path: Path):
        pq_html = pyquery.PyQuery(filename=str(html_path))
        pq_img = pq_html("img")

        # 画像をすべてアップロードして、img要素のsrc属性値を annofab urlに変更する
        for img_elm in pq_img:
            src_value: str = img_elm.attrib.get("src")
            if src_value is None:
                continue

            if src_value.startswith("http:") or src_value.startswith("https:") or src_value.startswith("data:"):
                continue

            if src_value[0] == "/":
                img_path = Path(src_value)
            else:
                img_path = html_path.parent / src_value

            if img_path.exists():
                image_id = str(uuid.uuid4())
                img_url = self.service.wrapper.upload_instruction_image(project_id, image_id, str(img_path))

                logger.debug(f"image uploaded. file={img_path}, instruction_image_url={img_url}")
                img_elm.attrib["src"] = img_url
                time.sleep(1)

            else:
                logger.warning(f"image does not exist. path={img_path}")

        # 作業ガイドの更新(body element)
        html_data = pq_html("body").html()
        self.update_instruction(project_id, html_data)
        logger.info("作業ガイドを更新しました。")

    def update_instruction(self, project_id: str, html_data: str):
        histories, _ = self.service.api.get_instruction_history(project_id)
        if len(histories) > 0:
            last_updated_datetime = histories[0]["updated_datetime"]
        else:
            last_updated_datetime = None

        request_body = {"html": html_data, "last_updated_datetime": last_updated_datetime}
        self.service.api.put_instruction(project_id, request_body=request_body)

    def main(self):
        args = self.args

        self.upload_html_to_instruction(args.project_id, Path(args.html))


def main(args):
    service = build_annofabapi_resource_and_login(args)
    facade = AnnofabApiFacade(service)
    UploadInstruction(service, facade, args).main()


def parse_args(parser: argparse.ArgumentParser):
    argument_parser = ArgumentParser(parser)

    argument_parser.add_project_id()

    # TODO body要素内のみ？
    parser.add_argument("--html", type=str, required=True, help="作業ガイドとして登録するHTMLファイルのパスを指定します。")

    parser.set_defaults(subcommand_func=main)


def add_parser(subparsers: argparse._SubParsersAction):
    subcommand_name = "upload"
    subcommand_help = "HTMLファイルを作業ガイドとして登録します。"
    description = (
        "HTMLファイルを作業ガイドとして登録します。" + "img要素のsrc属性がローカルの画像を参照している場合（http, https, dataスキーマが付与されていない）、" "画像もアップロードします。"
    )
    epilog = "チェッカーまたはオーナロールを持つユーザで実行してください。"
    parser = annofabcli.common.cli.add_parser(subparsers, subcommand_name, subcommand_help, description, epilog=epilog)
    parse_args(parser)



import argparse
import logging
import urllib.parse
from typing import Any, Dict, List, Optional

import annofabapi
from annofabapi.models import InputData, Task

import annofabcli
from annofabcli import AnnofabApiFacade
from annofabcli.common.cli import AbstractCommandLineInterface, ArgumentParser, build_annofabapi_resource_and_login
from annofabcli.common.enums import FormatArgument
from annofabcli.common.visualize import AddProps

logger = logging.getLogger(__name__)


class ListInputData(AbstractCommandLineInterface):
    """
    入力データの一覧を表示する
    """

    #: 入力データIDの平均長さ
    average_input_data_id_length: int = 36

    def __init__(self, service: annofabapi.Resource, facade: AnnofabApiFacade, args: argparse.Namespace):
        super().__init__(service, facade, args)
        self.visualize = AddProps(self.service, args.project_id)
        self.average_input_data_id_length = args.averate_input_data_id_length

    @staticmethod
    def _find_task_id_list(task_list: List[Task], input_data_id: str) -> List[str]:
        """
        タスク一覧から、該当のinput_data_idを持つtask_id_listを返す。
        """
        task_id_list = []
        for task in task_list:
            if input_data_id in task["input_data_id_list"]:
                task_id_list.append(task["task_id"])
        return task_id_list

    @annofabcli.utils.allow_404_error
    def get_input_data(self, project_id: str, input_data_id: str) -> InputData:
        input_data, _ = self.service.api.get_input_data(project_id, input_data_id)
        return input_data

    def get_input_data_from_input_data_id(self, project_id: str, input_data_id_list: List[str]) -> List[InputData]:
        input_data_list = []
        logger.debug(f"{len(input_data_id_list)}件の入力データを取得します。")
        for index, input_data_id in enumerate(input_data_id_list):
            if (index + 1) % 100 == 0:
                logger.debug(f"{index+1} 件目の入力データを取得します。")

            input_data = self.get_input_data(project_id, input_data_id)
            if input_data is not None:
                input_data_list.append(input_data)
            else:
                logger.warning(f"入力データ '{input_data_id}' は見つかりませんでした。")

        return input_data_list

    def add_details_to_input_data_list(self, project_id: str, input_data_list: List[InputData]) -> List[InputData]:
        """
        `input_data_list`に詳細情報（どのタスクに使われているか）を付与する。

        Args:
            project_id:
            input_data_list: 入力データList(In/Out)

        Returns:

        """
        if len(input_data_list) == 0:
            return input_data_list

        logger.debug(f"入力データ一覧の件数: {len(input_data_list)}")

        # AWS CloudFrontのURLの上限が8,192byte
        # https://docs.aws.amazon.com/AmazonCloudFront/latest/DeveloperGuide/cloudfront-limits.html
        MAX_URL_QUERY_LENGTH = 8000  # input_dat_ids部分のURLクエリの最大値
        average_input_data_id_length = self.average_input_data_id_length + 1  # カンマの分だけ長さを増やす
        chunk_size = MAX_URL_QUERY_LENGTH // average_input_data_id_length
        initial_index = 0
        while True:
            sub_input_data_list = input_data_list[initial_index : initial_index + chunk_size]
            sub_input_data_id_list = [e["input_data_id"] for e in sub_input_data_list]
            str_input_data_id_list = ",".join(sub_input_data_id_list)
            encoded_input_data_id_list = urllib.parse.quote(str_input_data_id_list)
            if len(encoded_input_data_id_list) > MAX_URL_QUERY_LENGTH:
                differential_length = len(encoded_input_data_id_list) - MAX_URL_QUERY_LENGTH
                decreasing_size = (differential_length // average_input_data_id_length) + 1
                logger.debug(
                    f"chunk_sizeを {chunk_size} から、{chunk_size - decreasing_size} に減らした. "
                    f"len(encoded_input_data_id_list) = {len(encoded_input_data_id_list)}"
                )
                chunk_size = chunk_size - decreasing_size
                if chunk_size <= 0:
                    chunk_size = 1

                continue

            logger.debug(f"input_data_list[{initial_index}:{initial_index+chunk_size}] を使用しているタスクを取得する。")
            task_list = self.service.wrapper.get_all_tasks(
                project_id, query_params={"input_data_ids": str_input_data_id_list}
            )

            for input_data in sub_input_data_list:
                # input_data_idで絞り込んでいるが、大文字小文字を区別しない。
                # したがって、確認のため `_find_task_id_list`を実行する
                task_id_list = self._find_task_id_list(task_list, input_data["input_data_id"])
                self.visualize.add_properties_to_input_data(input_data, task_id_list)

            initial_index = initial_index + chunk_size
            if initial_index >= len(input_data_list):
                break

        return input_data_list

    def get_input_data_list(
        self,
        project_id: str,
        input_data_id_list: Optional[List[str]] = None,
        input_data_query: Optional[Dict[str, Any]] = None,
        add_details: bool = False,
    ) -> List[InputData]:
        """
        入力データ一覧を取得する。
        """
        if input_data_id_list is not None:
            input_data_list = self.get_input_data_from_input_data_id(project_id, input_data_id_list)
        else:
            logger.debug(f"input_data_query: {input_data_query}")
            input_data_list = self.service.wrapper.get_all_input_data_list(project_id, query_params=input_data_query)

        # 詳細な情報を追加する
        if add_details:
            self.add_details_to_input_data_list(project_id, input_data_list)

        return input_data_list

    def print_input_data(
        self,
        project_id: str,
        input_data_query: Optional[Dict[str, Any]] = None,
        input_data_id_list: Optional[List[str]] = None,
        add_details: bool = False,
    ):
        """
        入力データ一覧を出力する

        Args:
            project_id: 対象のproject_id
            input_data_query: 入力データの検索クエリ
            input_data_id: 入力データID
            add_details: 詳細情報を表示する

        """

        super().validate_project(project_id, project_member_roles=None)

        input_data_list = self.get_input_data_list(
            project_id,
            input_data_query=input_data_query,
            input_data_id_list=input_data_id_list,
            add_details=add_details,
        )
        logger.info(f"入力データ一覧の件数: {len(input_data_list)}")
        if len(input_data_list) == 10000:
            logger.warning("入力データ一覧は10,000件で打ち切られている可能性があります。")

        if len(input_data_list) > 0:
            self.print_according_to_format(input_data_list)
        else:
            logger.info(f"入力データの件数が0件のため、出力しません。")

    def main(self):
        args = self.args
        input_data_id_list = annofabcli.common.cli.get_list_from_args(args.input_data_id)
        if len(input_data_id_list) == 0:
            input_data_id_list = None

        input_data_query = annofabcli.common.cli.get_json_from_args(args.input_data_query)
        self.print_input_data(
            args.project_id,
            input_data_id_list=input_data_id_list,
            input_data_query=input_data_query,
            add_details=args.add_details,
        )


def main(args):
    service = build_annofabapi_resource_and_login(args)
    facade = AnnofabApiFacade(service)
    ListInputData(service, facade, args).main()


def parse_args(parser: argparse.ArgumentParser):
    argument_parser = ArgumentParser(parser)

    argument_parser.add_project_id()

    query_group = parser.add_mutually_exclusive_group()

    query_group.add_argument(
        "-iq",
        "--input_data_query",
        type=str,
        help="入力データの検索クエリをJSON形式で指定します。"
        "`file://`を先頭に付けると、JSON形式のファイルを指定できます。"
        "クエリのフォーマットは、[getInputDataList API](https://annofab.com/docs/api/#operation/getInputDataList)のクエリパラメータと同じです。"
        "ただし `page`, `limit`キーは指定できません。",
    )

    query_group.add_argument(
        "-i",
        "--input_data_id",
        type=str,
        nargs="+",
        help="対象のinput_data_idを指定します。`--input_data_query`引数とは同時に指定できません。"
        "`file://`を先頭に付けると、input_data_idの一覧が記載されたファイルを指定できます。",
    )

    parser.add_argument("--add_details", action="store_true", help="入力データの詳細情報を表示します（`parent_task_id_list`）")

    parser.add_argument(
        "--averate_input_data_id_length",
        type=int,
        default=36,
        help=("入力データIDの平均長さを指定します。`add_details`がTrueのときのみ有効です。" "デフォルトはUUIDv4の長さです。" "この値を元にして、タスク一括取得APIの実行回数を決めます。"),
    )

    argument_parser.add_format(
        choices=[
            FormatArgument.CSV,
            FormatArgument.JSON,
            FormatArgument.PRETTY_JSON,
            FormatArgument.INPUT_DATA_ID_LIST,
        ],
        default=FormatArgument.CSV,
    )
    argument_parser.add_output()
    argument_parser.add_csv_format()

    argument_parser.add_query()
    parser.set_defaults(subcommand_func=main)


def add_parser(subparsers: argparse._SubParsersAction):
    subcommand_name = "list"
    subcommand_help = "入力データ一覧を出力します。"
    description = "入力データ一覧を出力します。AnnoFabの制約上、10,000件までしか出力されません。"

    parser = annofabcli.common.cli.add_parser(subparsers, subcommand_name, subcommand_help, description)
    parse_args(parser)



import argparse
import logging

import annofabapi
from annofabapi.models import JobStatus, JobType

import annofabcli
import annofabcli.common.cli
from annofabcli import AnnofabApiFacade
from annofabcli.common.cli import (
    AbstractCommandLineInterface,
    ArgumentParser,
    build_annofabapi_resource_and_login,
    get_json_from_args,
    get_wait_options_from_args,
)
from annofabcli.common.dataclasses import WaitOptions

logger = logging.getLogger(__name__)


class WaitJobMain:
    def __init__(self, service: annofabapi.Resource):
        self.service = service
        self.facade = AnnofabApiFacade(service)

    def is_job_progress(self, project_id: str, job_type: JobType) -> bool:
        job_list = self.service.api.get_project_job(project_id, query_params={"type": job_type.value})[0]["list"]
        if len(job_list) > 0:
            if job_list[0]["job_status"] == JobStatus.PROGRESS.value:
                return True

        return False

    def wait_job(self, project_id: str, job_type: JobType, wait_options: WaitOptions):
        if not self.is_job_progress(project_id, job_type):
            logger.info(f"job_type='{job_type.value}'の実行中のジョブはないので、終了します。")
            return

        MAX_WAIT_MINUTUE = wait_options.max_tries * wait_options.interval / 60
        logger.info(f"ダウンロード対象の最新化処理が完了するまで、最大{MAX_WAIT_MINUTUE}分間待ちます。")
        result = self.service.wrapper.wait_for_completion(
            project_id,
            job_type=job_type,
            job_access_interval=wait_options.interval,
            max_job_access=wait_options.max_tries,
        )
        if result:
            logger.info(f"job_type='{job_type.value}'のジョブが終了しました。")
        else:
            logger.info(f"job_type='{job_type.value}'のジョブが失敗したか、または {MAX_WAIT_MINUTUE} 分待っても処理が終了しませんでした。")


class WaitJob(AbstractCommandLineInterface):
    def main(self):
        args = self.args
        project_id = args.project_id
        job_type = JobType(args.job_type)

        DEFAULT_WAIT_OPTIONS = WaitOptions(interval=60, max_tries=360)
        wait_options = get_wait_options_from_args(get_json_from_args(args.wait_options), DEFAULT_WAIT_OPTIONS)

        main_obj = WaitJobMain(self.service)
        main_obj.wait_job(project_id, job_type=job_type, wait_options=wait_options)


def main(args):
    service = build_annofabapi_resource_and_login(args)
    facade = AnnofabApiFacade(service)
    WaitJob(service, facade, args).main()


def parse_args(parser: argparse.ArgumentParser):
    argument_parser = ArgumentParser(parser)

    argument_parser.add_project_id()

    job_choices = [e.value for e in JobType]
    parser.add_argument("--job_type", type=str, choices=job_choices, required=True, help="ジョブタイプを指定します。")

    parser.add_argument(
        "--wait_options",
        type=str,
        help="ジョブの終了を待つときのオプションをJSON形式で指定してください。"
        "`file://`を先頭に付けるとjsonファイルを指定できます。"
        'デフォルとは`{"interval":60, "max_tries":360}` です。'
        "`interval`:ジョブが完了したかを問い合わせる間隔[秒], "
        "`max_tires`:ジョブが完了したかの問い合わせを最大何回行うか。",
    )

    parser.set_defaults(subcommand_func=main)


def add_parser(subparsers: argparse._SubParsersAction):
    subcommand_name = "wait"
    subcommand_help = "ジョブの終了を待ちます。"
    description = "ジョブの終了を待ちます。"
    parser = annofabcli.common.cli.add_parser(subparsers, subcommand_name, subcommand_help, description)
    parse_args(parser)



import argparse
import logging
import sys
from typing import Any, Dict, List, Optional

import requests
from annofabapi.models import ProjectMember, ProjectMemberRole, ProjectMemberStatus

import annofabcli
from annofabcli import AnnofabApiFacade
from annofabcli.common.cli import AbstractCommandLineInterface, ArgumentParser, build_annofabapi_resource_and_login
from annofabcli.project_member.put_project_members import PutProjectMembers

logger = logging.getLogger(__name__)


class ChangeProjectMembers(AbstractCommandLineInterface):
    """
    プロジェクトメンバのメンバ情報を更新する。
    """

    def put_project_member(
        self,
        project_id: str,
        user_id: str,
        old_member: ProjectMember,
        member_role: Optional[ProjectMemberRole] = None,
        member_info: Optional[Dict[str, Any]] = None,
    ) -> ProjectMember:
        """
        1人のプロジェクトメンバを変更する。

        Args:
            project_id:
            user_id:
            old_member: 変更前のメンバ情報
            member_role: 変更後のロール。Noneの場合、変更しない。
            member_info: 変更後のメンバ情報。Noneの場合、変更しない。また、対象のキーが存在しない場合は、そのキーに対して変更しない。

        Returns:

        """

        def get_value(key):
            if member_info is None:
                return old_member[key]

            if key not in member_info:
                return old_member[key]

            return member_info[key]

        str_member_role = member_role.value if member_role is not None else old_member["member_role"]

        request_body = {
            "member_status": ProjectMemberStatus.ACTIVE.value,
            "member_role": str_member_role,
            "sampling_inspection_rate": get_value("sampling_inspection_rate"),
            "sampling_acceptance_rate": get_value("sampling_acceptance_rate"),
            "last_updated_datetime": old_member["updated_datetime"],
        }

        updated_project_member = self.service.api.put_project_member(project_id, user_id, request_body=request_body)[0]
        return updated_project_member

    def change_project_members(
        self,
        project_id: str,
        user_id_list,
        member_role: Optional[ProjectMemberRole] = None,
        member_info: Optional[Dict[str, Any]] = None,
    ):
        """
        プロジェクトメンバのメンバ情報を更新する。

        Args:
            project_id: プロジェクトメンバの登録先のプロジェクトのプロジェクトID
            user_id_list: 変更対象のプロジェクトメンバのuser_id
            member_role: メンバに対してして設定するロール。
            member_info: プロジェクトメンバに対して設定するメンバ情報

        """

        super().validate_project(project_id, [ProjectMemberRole.OWNER])

        old_project_members = self.service.wrapper.get_all_project_members(project_id)
        project_title = self.facade.get_project_title(project_id)

        count_invite_members = 0
        # プロジェクトメンバを登録
        logger.info(f"{project_title} に、{len(user_id_list)} 件のプロジェクトメンバの情報を変更します。")
        for user_id in user_id_list:
            if user_id == self.service.api.login_user_id:
                logger.warning(f"ユーザ '{user_id}'は自分自身なので、変更できません。")
                continue

            old_member = PutProjectMembers.find_member(old_project_members, user_id)
            if old_member is None:
                logger.warning(f"ユーザ '{user_id}' は、プロジェクトメンバでないため変更できませんでした。")
                continue

            message_for_confirm = f"ユーザ '{user_id}'のプロジェクトメンバ情報を変更しますか？"
            if not self.confirm_processing(message_for_confirm):
                continue

            # メンバを登録
            try:
                self.put_project_member(project_id, user_id, member_info=member_info, old_member=old_member)
                logger.debug(
                    f"user_id = {user_id} のプロジェクトメンバ情報を変更しました。member_role={member_role}, member_info={member_info}"
                )
                count_invite_members += 1

            except requests.exceptions.HTTPError as e:
                logger.warning(e)
                logger.warning(f"プロジェクトメンバの登録に失敗しました。user_id={user_id}")

        logger.info(f"{project_title} に、{count_invite_members} / {len(user_id_list)} 件のプロジェクトメンバを変更しました。")

    def get_all_user_id_list_except_myself(self, project_id: str) -> List[str]:
        """自分自身を除いた、すべてのプロジェクトメンバを取得する"""
        member_list = self.service.wrapper.get_all_project_members(project_id)
        return [e["user_id"] for e in member_list if e["user_id"] != self.service.api.login_user_id]

    @staticmethod
    def validate(args: argparse.Namespace, member_info: Optional[Dict[str, Any]] = None) -> bool:
        COMMON_MESSAGE = "annofabcli inspection_comment list: error:"
        if args.role is None and args.member_info is None:
            print(f"{COMMON_MESSAGE} argument `--role`または`--member_info`のどちらかは、必ず指定してください。", file=sys.stderr)
            return False

        elif member_info is not None and not ChangeProjectMembers.validate_member_info(member_info):
            print(f"{COMMON_MESSAGE} argument --member_info: 有効なキーが１つも指定されていません。", file=sys.stderr)
            return False

        else:
            return True

    @staticmethod
    def validate_member_info(member_info: Dict[str, Any]):
        KEYS = ["sampling_inspection_rate", "sampling_acceptance_rate"]
        return any(k in member_info for k in KEYS)

    def main(self):
        args = self.args
        project_id = args.project_id
        if args.all_user:
            user_id_list = self.get_all_user_id_list_except_myself(project_id)
        else:
            user_id_list = annofabcli.common.cli.get_list_from_args(args.user_id)

        member_info = annofabcli.common.cli.get_json_from_args(args.member_info)
        member_role = ProjectMemberRole(args.role) if args.role is not None else None

        if not self.validate(args, member_info):
            return

        self.change_project_members(
            args.project_id, user_id_list=user_id_list, member_role=member_role, member_info=member_info
        )


def main(args):
    service = build_annofabapi_resource_and_login(args)
    facade = AnnofabApiFacade(service)
    ChangeProjectMembers(service, facade, args).main()


def parse_args(parser: argparse.ArgumentParser):
    argument_parser = ArgumentParser(parser)
    role_choices = [e.value for e in ProjectMemberRole]

    argument_parser.add_project_id()

    user_group = parser.add_mutually_exclusive_group(required=True)
    user_group.add_argument(
        "-u",
        "--user_id",
        type=str,
        nargs="+",
        help="変更するプロジェクトメンバのuser_idを指定してください。" "`file://`を先頭に付けると、一覧が記載されたファイルを指定できます。",
    )
    user_group.add_argument("--all_user", action="store_true", help="自分以外のすべてのプロジェクトメンバを変更します。")

    parser.add_argument(
        "--role", type=str, choices=role_choices, help="プロジェクトメンバにユーザに割り当てるロールを指定します。指定しない場合は、ロールは変更されません。"
    )

    parser.add_argument(
        "--member_info",
        type=str,
        required=True,
        help="プロジェクトメンバに対して設定するメンバ情報を、JSON形式で指定します。`file://`を先頭に付けると、JSON形式のファイルを指定できます。 "
        "以下のキーが指定可能です。sampling_inspection_rate, sampling_acceptance_rate, "
        "未設定にする場合は、値にnullを指定してください。"
        "詳細は https://annofab.com/docs/api/#operation/putProjectMember を参照ください。 ",
    )

    parser.set_defaults(subcommand_func=main)


def add_parser(subparsers: argparse._SubParsersAction):
    subcommand_name = "change"
    subcommand_help = "プロジェクトメンバを変更します。"
    description = "複数のプロジェクトメンバに対して、メンバ情報を変更します。ただし、自分自身は変更できません。"
    epilog = "オーナロールを持つユーザで実行してください。"

    parser = annofabcli.common.cli.add_parser(subparsers, subcommand_name, subcommand_help, description, epilog=epilog)
    parse_args(parser)



import argparse
import logging
import multiprocessing
import sys
from dataclasses import dataclass
from functools import partial
from typing import List, Optional, Tuple

import annofabapi
import requests
from annofabapi.dataclass.task import Task
from annofabapi.models import ProjectMemberRole

import annofabcli
import annofabcli.common.cli
from annofabcli import AnnofabApiFacade
from annofabcli.common.cli import (
    AbstracCommandCinfirmInterface,
    AbstractCommandLineInterface,
    ArgumentParser,
    build_annofabapi_resource_and_login,
)
from annofabcli.common.facade import TaskQuery, match_task_with_query

logger = logging.getLogger(__name__)


@dataclass(frozen=True)
class User:
    account_id: str
    user_id: str
    username: str


class CancelAcceptanceMain(AbstracCommandCinfirmInterface):
    def __init__(self, service: annofabapi.Resource, all_yes: bool = False):
        self.service = service
        self.facade = AnnofabApiFacade(service)
        AbstracCommandCinfirmInterface.__init__(self, all_yes)

    def cancel_acceptance_for_task(
        self,
        project_id: str,
        task_id: str,
        acceptor: Optional[User] = None,
        assign_last_acceptor: bool = True,
        task_query: Optional[TaskQuery] = None,
        task_index: Optional[int] = None,
    ) -> bool:
        def get_user_id(user: Optional[User]) -> Optional[str]:
            return user.user_id if user is not None else None

        logging_prefix = f"{task_index + 1} 件目" if task_index is not None else ""

        try:
            task, _ = self.service.api.get_task(project_id, task_id)

            if task["status"] != "complete":
                logger.warning(
                    f"{logging_prefix}: task_id = {task_id} は受入完了でありません。"
                    f"status = {task['status']}, phase={task['phase']}"
                )
                return False

            actual_acceptor: Optional[User] = None
            if assign_last_acceptor:
                acceptor_account_id = task["account_id"]
                if acceptor_account_id is not None:
                    member = self.facade.get_project_member_from_account_id(project_id, acceptor_account_id)
                    if member is not None:
                        actual_acceptor = User(
                            account_id=member["account_id"], user_id=member["user_id"], username=member["username"]
                        )
                    else:
                        logger.warning(f"account_id='{acceptor_account_id}' であるユーザは見つかりませんでした。")
                        return False
            else:
                actual_acceptor = acceptor

            if not match_task_with_query(Task.from_dict(task), task_query):
                logger.debug(f"{logging_prefix} : task_id = {task_id} : TaskQueryの条件にマッチしないため、スキップします。")
                return False

            if not self.confirm_processing(
                f"{logging_prefix}: task_id = {task_id} のタスクの受入を取り消しますか？ "
                f"user_id = '{get_user_id(actual_acceptor)}' のユーザに割り当てます。"
            ):
                return False

            logger.debug(
                f"{logging_prefix}: task_id = {task_id} のタスクの受入を取り消し、"
                f"user_id = '{get_user_id(actual_acceptor)}' のユーザに割り当てます。"
            )
            request_body = {
                "status": "not_started",
                "account_id": actual_acceptor.account_id if actual_acceptor is not None else None,
                "last_updated_datetime": task["updated_datetime"],
            }
            self.service.api.operate_task(project_id, task_id, request_body=request_body)
            logger.info(f"{logging_prefix} : task_id = {task_id} の受け入れ取り消しが成功しました。")
            return True

        except requests.exceptions.HTTPError as e:
            logger.warning(f"{logging_prefix} : task_id = {task_id} の受け入れ取り消しに失敗しました。")
            logger.warning(e)
            return False

    def cancel_acceptance_for_wrapper(
        self,
        tpl: Tuple[int, str],
        project_id: str,
        acceptor: Optional[User] = None,
        assign_last_acceptor: bool = True,
        task_query: Optional[TaskQuery] = None,
    ) -> bool:
        task_index, task_id = tpl
        return self.cancel_acceptance_for_task(
            project_id=project_id,
            task_id=task_id,
            acceptor=acceptor,
            assign_last_acceptor=assign_last_acceptor,
            task_query=task_query,
            task_index=task_index,
        )

    def cancel_acceptance_for_task_list(
        self,
        project_id: str,
        task_id_list: List[str],
        acceptor_user_id: Optional[str] = None,
        assign_last_acceptor: bool = True,
        task_query: Optional[TaskQuery] = None,
        parallelism: Optional[int] = None,
    ):
        """
        タスクを受入取り消しする

        Args:
            project_id:
            task_id_list: 受け入れ取り消しするtask_id_list
            acceptor_user_id: 再度受入を担当させたいユーザのuser_id. Noneならば、未割り当てにする
            assign_last_acceptor: trueなら最後の受入担当者に割り当てる
        """
        logger.info(f"受け入れを取り消すタスク数: {len(task_id_list)}")

        if task_query is not None:
            task_query = self.facade.set_account_id_of_task_query(project_id, task_query)

        acceptor: Optional[User] = None
        if acceptor_user_id is not None:
            member = self.facade.get_project_member_from_user_id(project_id, acceptor_user_id)
            if member is not None:
                acceptor = User(account_id=member["account_id"], user_id=member["user_id"], username=member["username"])
            else:
                raise RuntimeError(f"user_id='{acceptor_user_id}' であるプロジェクトメンバは見つかりませんでした。")

        success_count = 0
        if parallelism is not None:
            partial_func = partial(
                self.cancel_acceptance_for_wrapper,
                project_id=project_id,
                acceptor=acceptor,
                assign_last_acceptor=assign_last_acceptor,
                task_query=task_query,
            )
            with multiprocessing.Pool(parallelism) as pool:
                result_bool_list = pool.map(partial_func, enumerate(task_id_list))
                success_count = len([e for e in result_bool_list if e])

        else:
            # 逐次処理
            success_count = 0
            for task_index, task_id in enumerate(task_id_list):
                result = self.cancel_acceptance_for_task(
                    project_id,
                    task_id,
                    task_index=task_index,
                    acceptor=acceptor,
                    assign_last_acceptor=assign_last_acceptor,
                    task_query=task_query,
                )
                if result:
                    success_count += 1

        logger.info(f"{success_count} / {len(task_id_list)} 件 受け入れ取り消しに成功しました。")


class CancelAcceptance(AbstractCommandLineInterface):
    @staticmethod
    def validate(args: argparse.Namespace) -> bool:
        COMMON_MESSAGE = "annofabcli task cancel_acceptance: error:"

        if args.parallelism is not None and not args.yes:
            print(
                f"{COMMON_MESSAGE} argument --parallelism: '--parallelism'を指定するときは、必ず'--yes'を指定してください。",
                file=sys.stderr,
            )
            return False

        return True

    def main(self):
        args = self.args
        if not self.validate(args):
            return

        task_id_list = annofabcli.common.cli.get_list_from_args(args.task_id)

        assign_last_acceptor = not args.not_assign and args.assigned_acceptor_user_id is None

        dict_task_query = annofabcli.common.cli.get_json_from_args(args.task_query)
        task_query: Optional[TaskQuery] = TaskQuery.from_dict(dict_task_query) if dict_task_query is not None else None

        super().validate_project(args.project_id, [ProjectMemberRole.OWNER])
        main_obj = CancelAcceptanceMain(self.service, all_yes=args.yes)
        main_obj.cancel_acceptance_for_task_list(
            args.project_id,
            task_id_list,
            acceptor_user_id=args.assigned_acceptor_user_id,
            assign_last_acceptor=assign_last_acceptor,
            task_query=task_query,
            parallelism=args.parallelism,
        )


def main(args: argparse.Namespace):
    service = build_annofabapi_resource_and_login(args)
    facade = AnnofabApiFacade(service)
    CancelAcceptance(service, facade, args).main()


def parse_args(parser: argparse.ArgumentParser):
    argument_parser = ArgumentParser(parser)

    parser.add_argument("-p", "--project_id", type=str, required=True, help="対象のプロジェクトのproject_idを指定します。")

    parser.add_argument(
        "-t",
        "--task_id",
        type=str,
        required=True,
        nargs="+",
        help="対象のタスクのtask_idを指定します。`file://`を先頭に付けると、task_idの一覧が記載されたファイルを指定できます。",
    )

    # 受入取消後のタスクの担当者の割当に関して
    assign_group = parser.add_mutually_exclusive_group()

    assign_group.add_argument(
        "--not_assign", action="store_true", help="受入を取り消した後のタスクの担当者を未割り当てにします。" "指定しない場合は、最後の受入phaseの担当者が割り当てます。"
    )

    assign_group.add_argument(
        "--assigned_acceptor_user_id",
        type=str,
        help="受入を取り消した後に割り当てる受入作業者のuser_idを指定します。" "指定しない場合は、最後の受入phaseの担当者が割り当てます。",
    )

    argument_parser.add_task_query()

    parser.add_argument(
        "--parallelism", type=int, help="使用するプロセス数（並列度）を指定してください。指定する場合は必ず'--yes'を指定してください。指定しない場合は、逐次的に処理します。"
    )

    parser.set_defaults(subcommand_func=main)


def add_parser(subparsers: argparse._SubParsersAction):
    subcommand_name = "cancel_acceptance"
    subcommand_help = "受入が完了したタスクに対して、受入を取り消します。"
    description = "受入が完了したタスクに対して、受入を取り消します。"
    epilog = "オーナロールを持つユーザで実行してください。"

    parser = annofabcli.common.cli.add_parser(subparsers, subcommand_name, subcommand_help, description, epilog=epilog)
    parse_args(parser)



# pylint: disable=too-many-lines
import logging
from enum import Enum
from pathlib import Path
from typing import Dict, List, Optional

import bokeh
import bokeh.layouts
import bokeh.palettes
import dateutil
import pandas
from bokeh.core.properties import Color
from bokeh.models import DataRange1d, HoverTool, LinearAxis
from bokeh.plotting import ColumnDataSource, figure

logger = logging.getLogger(__name__)


class OutputTarget(Enum):
    ANNOTATION = "annotation"
    INPUT_DATA = "input_data"
    TASK = "task"


class LineGraph:
    """
    グラフを出力するクラス
    """

    #############################################
    # Field
    #############################################
    my_palette = bokeh.palettes.Category20[20]
    my_small_palette = bokeh.palettes.Category10[10]

    #############################################
    # Private
    #############################################

    def __init__(self, outdir: str):
        self.line_graph_outdir = outdir
        Path(self.line_graph_outdir).mkdir(exist_ok=True, parents=True)

    @staticmethod
    def _create_hover_tool(tool_tip_items: Optional[List[str]] = None) -> HoverTool:
        """
        HoverTool用のオブジェクトを生成する。
        Returns:

        """
        if tool_tip_items is None:
            tool_tip_items = []

        detail_tooltips = [(e, "@{%s}" % e) for e in tool_tip_items]

        hover_tool = HoverTool(tooltips=[("index", "$index"), ("(x,y)", "($x, $y)")] + detail_tooltips)
        return hover_tool

    @staticmethod
    def _plot_line_and_circle(
        fig: bokeh.plotting.Figure,
        source: ColumnDataSource,
        x_column_name: str,
        y_column_name: str,
        legend_label: str,
        color: Color,
        **kwargs,
    ) -> None:
        """
        線を引いて、プロットした部分に丸を付ける。

        Args:
            fig:
            source:
            x_column_name: sourceに対応するX軸の列名
            y_column_name: sourceに対応するY軸の列名
            legend_label:
            color: 線と点の色

        """

        fig.line(
            x=x_column_name,
            y=y_column_name,
            source=source,
            legend_label=legend_label,
            line_color=color,
            line_width=1,
            muted_alpha=0,
            muted_color=color,
            **kwargs,
        )
        fig.circle(
            x=x_column_name,
            y=y_column_name,
            source=source,
            legend_label=legend_label,
            muted_alpha=0.0,
            muted_color=color,
            color=color,
            **kwargs,
        )

    @staticmethod
    def _plot_moving_average(
        fig: bokeh.plotting.Figure,
        source: ColumnDataSource,
        x_column_name: str,
        y_column_name: str,
        legend_label: str,
        color: Color,
        **kwargs,
    ) -> None:
        """
        移動平均用にプロットする

        Args:
            fig:
            source:
            x_column_name: sourceに対応するX軸の列名
            y_column_name: sourceに対応するY軸の列名
            legend_label:
            color: 線と点の色

        """

        fig.line(
            x=x_column_name,
            y=y_column_name,
            source=source,
            legend_label=legend_label,
            line_color=color,
            line_width=1,
            line_dash="dashed",
            line_alpha=0.6,
            muted_alpha=0,
            muted_color=color,
            **kwargs,
        )

    @staticmethod
    def _set_legend(fig: bokeh.plotting.Figure, hover_tool: HoverTool) -> None:
        """
        凡例の設定。

        Args:
            fig:
            hover_tool:
        """
        fig.add_tools(hover_tool)
        fig.legend.location = "top_left"
        fig.legend.click_policy = "mute"
        if len(fig.legend) > 0:
            legend = fig.legend[0]
            fig.add_layout(legend, "left")

    def create_user_id_list(
        self,
        df: pandas.DataFrame,
        user_id_column: str,
        datetime_column: str,
        arg_user_id_list: Optional[List[str]] = None,
    ) -> List[str]:
        """
        グラフに表示するユーザのuser_idを生成する。

        Args:
            df:
            user_id_column: user_idが格納されている列の名前
            datetime_column: 日時列の降順でソートする
            arg_first_annotation_user_id_list: 指定されたuser_idのList

        Returns:
            グラフに表示するユーザのuser_idのList

        """
        max_user_length = len(self.my_palette)
        tmp_user_id_list: List[str] = (
            df.sort_values(by=datetime_column, ascending=False)[user_id_column].dropna().unique().tolist()
        )

        if arg_user_id_list is None or len(arg_user_id_list) == 0:
            user_id_list = tmp_user_id_list
        else:
            user_id_list = [user_id for user_id in tmp_user_id_list if user_id in arg_user_id_list]

        if len(user_id_list) > max_user_length:
            logger.info(f"表示対象のuser_idの数が多いため、先頭から{max_user_length}個のみグラフ化します")

        return user_id_list[0:max_user_length]

    def write_productivity_line_graph_for_annotator(
        self,
        df: pandas.DataFrame,
        first_annotation_user_id_list: Optional[List[str]] = None,
    ):
        """
        生産性を教師付作業者ごとにプロットする。

        Args:
            df:
            first_annotation_user_id_list:

        Returns:

        """

        def write_cumulative_graph(fig_info_list: List[Dict[str, str]], html_title: str):
            """

            Args:
                fig_info_list:
                html_title:

            Returns:

            """
            output_file = f"{self.line_graph_outdir}/{html_title}.html"
            logger.debug(f"{output_file} を出力します。")

            figs: List[bokeh.plotting.Figure] = []
            for fig_info in fig_info_list:
                figs.append(
                    figure(
                        plot_width=1200,
                        plot_height=600,
                        title=fig_info["title"],
                        x_axis_label=fig_info["x_axis_label"],
                        x_axis_type="datetime",
                        y_axis_label=fig_info["y_axis_label"],
                    )
                )

            for user_index, user_id in enumerate(user_id_list):
                filtered_df = df[df["first_annotation_user_id"] == user_id]
                if filtered_df.empty:
                    logger.debug(f"dataframe is empty. user_id = {user_id}")
                    continue

                source = ColumnDataSource(data=filtered_df)
                color = self.my_palette[user_index]
                username = filtered_df.iloc[0]["first_annotation_username"]

                for fig, fig_info in zip(figs, fig_info_list):
                    self._plot_line_and_circle(
                        fig,
                        x_column_name=fig_info["x"],
                        y_column_name=fig_info["y"],
                        source=source,
                        legend_label=username,
                        color=color,
                    )

            hover_tool = self._create_hover_tool(tooltip_item)
            for fig in figs:
                self._set_legend(fig, hover_tool)

            bokeh.plotting.reset_output()
            bokeh.plotting.output_file(output_file, title=html_title)
            bokeh.plotting.save(bokeh.layouts.column(figs))

        tooltip_item = [
            "first_annotation_user_id",
            "first_annotation_username",
            "first_annotation_started_date",
            "first_annotation_worktime_hour",
            "annotation_worktime_hour",
            "inspection_worktime_hour",
            "acceptance_worktime_hour",
            "annotation_count",
            "input_data_count",
            "task_count",
            "inspection_count",
        ]

        if len(df) == 0:
            logger.info("データが0件のため出力しない")
            return

        user_id_list = self.create_user_id_list(
            df,
            "first_annotation_user_id",
            datetime_column="first_annotation_started_date",
            arg_user_id_list=first_annotation_user_id_list,
        )
        logger.debug(f"教師付者用の折れ線グラフに表示する、教師付者のuser_id = {user_id_list}")

        df["date_first_annotation_started_date"] = df["first_annotation_started_date"].map(
            lambda e: dateutil.parser.parse(e).date()
        )

        fig_info_list_annotation_count = [
            dict(
                x="date_first_annotation_started_date",
                y="first_annotation_worktime_minute/annotation_count",
                title="アノテーションあたり教師付時間(1回目)の折れ線グラフ",
                x_axis_label="1回目の教師付開始日",
                y_axis_label="アノテーションあたり教師付時間(1回目)[min]",
            ),
            dict(
                x="date_first_annotation_started_date",
                y="annotation_worktime_minute/annotation_count",
                title="アノテーションあたり教師付時間の折れ線グラフ",
                x_axis_label="1回目の教師付開始日",
                y_axis_label="アノテーションあたり教師付時間[min]",
            ),
            dict(
                x="date_first_annotation_started_date",
                y="inspection_worktime_minute/annotation_count",
                title="アノテーションあたり検査時間の折れ線グラフ",
                x_axis_label="1回目の教師付開始日",
                y_axis_label="アノテーションあたり検査時間[min]",
            ),
            dict(
                x="date_first_annotation_started_date",
                y="acceptance_worktime_minute/annotation_count",
                title="アノテーションあたり受入時間の折れ線グラフ",
                x_axis_label="1回目の教師付開始日",
                y_axis_label="アノテーションあたり受入時間[min]",
            ),
            dict(
                x="date_first_annotation_started_date",
                y="inspection_count/annotation_count",
                title="アノテーションあたり検査コメント数の折れ線グラフ",
                x_axis_label="1回目の教師付開始日",
                y_axis_label="アノテーションあたり検査コメント数",
            ),
        ]
        write_cumulative_graph(
            fig_info_list_annotation_count,
            html_title="折れ線-横軸_教師付開始日-縦軸_アノテーションあたりの指標-教師付者用",
        )

        fig_info_list_input_data_count = [
            dict(
                x="date_first_annotation_started_date",
                y="first_annotation_worktime_minute/input_data_count",
                title="画像あたり教師付時間(1回目)の折れ線グラフ",
                x_axis_label="1回目の教師付開始日",
                y_axis_label="画像あたり教師付時間(1回目)[min]",
            ),
            dict(
                x="date_first_annotation_started_date",
                y="annotation_worktime_minute/input_data_count",
                title="画像あたり教師付時間の折れ線グラフ",
                x_axis_label="1回目の教師付開始日",
                y_axis_label="画像あたり教師付時間[min]",
            ),
            dict(
                x="date_first_annotation_started_date",
                y="inspection_worktime_minute/input_data_count",
                title="画像あたり検査時間の折れ線グラフ",
                x_axis_label="1回目の教師付開始日",
                y_axis_label="画像あたり検査時間[min]",
            ),
            dict(
                x="date_first_annotation_started_date",
                y="acceptance_worktime_minute/input_data_count",
                title="画像あたり受入時間の折れ線グラフ",
                x_axis_label="1回目の教師付開始日",
                y_axis_label="画像あたり受入時間[min]",
            ),
            dict(
                x="date_first_annotation_started_date",
                y="inspection_count/input_data_count",
                title="画像あたり検査コメント数の折れ線グラフ",
                x_axis_label="1回目の教師付開始日",
                y_axis_label="画像あたり検査コメント数",
            ),
        ]
        write_cumulative_graph(fig_info_list_input_data_count, html_title="折れ線-横軸_教師付開始日-縦軸_画像あたりの指標-教師付者用")

        fig_info_list_value = [
            dict(
                x="date_first_annotation_started_date",
                y="annotation_count",
                title="作業したアノテーション数の折れ線グラフ",
                x_axis_label="1回目の教師付開始日",
                y_axis_label="アノテーション数",
            ),
            dict(
                x="date_first_annotation_started_date",
                y="input_data_count",
                title="作業した画像数の折れ線グラフ",
                x_axis_label="1回目の教師付開始日",
                y_axis_label="画像数",
            ),
            dict(
                x="date_first_annotation_started_date",
                y="task_count",
                title="作業したタスク数の折れ線グラフ",
                x_axis_label="1回目の教師付開始日",
                y_axis_label="タスク数",
            ),
            dict(
                x="date_first_annotation_started_date",
                y="first_annotation_worktime_hour",
                title="教師付時間(1回目)の折れ線グラフ",
                x_axis_label="1回目の教師付開始日",
                y_axis_label="教師付時間(1回目)[hour]",
            ),
            dict(
                x="date_first_annotation_started_date",
                y="annotation_worktime_hour",
                title="教師付時間の折れ線グラフ",
                x_axis_label="1回目の教師付開始日",
                y_axis_label="教師付時間[hour]",
            ),
        ]
        write_cumulative_graph(fig_info_list_value, html_title="折れ線-横軸_教師付開始日-縦軸_指標-教師付者用")

    def write_productivity_line_graph_for_inspector(
        self,
        df: pandas.DataFrame,
        first_inspection_user_id_list: Optional[List[str]] = None,
    ):
        """
        日毎の生産性を検査作業者ごとにプロットする。

        Args:
            df:
            first_acceptance_user_id_list:

        Returns:

        """

        def write_graph(fig_info_list: List[Dict[str, str]], html_title: str):
            """

            Args:
                fig_info_list:
                html_title:

            Returns:

            """
            output_file = f"{self.line_graph_outdir}/{html_title}.html"
            logger.debug(f"{output_file} を出力します。")

            figs: List[bokeh.plotting.Figure] = []
            for fig_info in fig_info_list:
                figs.append(
                    figure(
                        plot_width=1200,
                        plot_height=600,
                        title=fig_info["title"],
                        x_axis_label=fig_info["x_axis_label"],
                        x_axis_type="datetime",
                        y_axis_label=fig_info["y_axis_label"],
                    )
                )

            for user_index, user_id in enumerate(user_id_list):
                filtered_df = df[df["first_inspection_user_id"] == user_id]
                if filtered_df.empty:
                    logger.debug(f"dataframe is empty. user_id = {user_id}")
                    continue

                source = ColumnDataSource(data=filtered_df)
                color = self.my_palette[user_index]
                username = filtered_df.iloc[0]["first_inspection_username"]

                for fig, fig_info in zip(figs, fig_info_list):
                    self._plot_line_and_circle(
                        fig,
                        x_column_name=fig_info["x"],
                        y_column_name=fig_info["y"],
                        source=source,
                        legend_label=username,
                        color=color,
                    )

            hover_tool = self._create_hover_tool(tooltip_item)
            for fig in figs:
                self._set_legend(fig, hover_tool)

            bokeh.plotting.reset_output()
            bokeh.plotting.output_file(output_file, title=html_title)
            bokeh.plotting.save(bokeh.layouts.column(figs))

        tooltip_item = [
            "first_inspection_user_id",
            "first_inspection_username",
            "first_inspection_started_date",
            "first_inspection_worktime_hour",
            "inspection_worktime_hour",
            "annotation_count",
            "input_data_count",
            "task_count",
            "inspection_count",
        ]

        if len(df) == 0:
            logger.info("データが0件のため出力しない")
            return

        user_id_list = self.create_user_id_list(
            df,
            "first_inspection_user_id",
            datetime_column="first_inspection_started_date",
            arg_user_id_list=first_inspection_user_id_list,
        )
        logger.debug(f"検査者用の折れ線グラフに表示する、受入者のuser_id = {user_id_list}")

        df["date_first_inspection_started_date"] = df["first_inspection_started_date"].map(
            lambda e: dateutil.parser.parse(e).date()
        )

        fig_info_list_inspection_count = [
            dict(
                x="date_first_inspection_started_date",
                y="first_inspection_worktime_minute/annotation_count",
                title="アノテーションあたり検査時間(1回目)の折れ線グラフ",
                x_axis_label="1回目の検査開始日",
                y_axis_label="アノテーションあたり検査時間(1回目)[min]",
            ),
            dict(
                x="date_first_inspection_started_date",
                y="inspection_worktime_minute/annotation_count",
                title="アノテーションあたり検査時間の折れ線グラフ",
                x_axis_label="1回目の検査開始日",
                y_axis_label="アノテーションあたり検査時間[min]",
            ),
        ]
        write_graph(
            fig_info_list_inspection_count,
            html_title="折れ線-横軸_検査開始日-縦軸_アノテーションあたりの指標-検査者用",
        )

        fig_info_list_input_data_count = [
            dict(
                x="date_first_inspection_started_date",
                y="first_inspection_worktime_minute/input_data_count",
                title="画像あたり検査時間(1回目)の折れ線グラフ",
                x_axis_label="1回目の検査開始日",
                y_axis_label="画像あたり検査時間(1回目)[min]",
            ),
            dict(
                x="date_first_inspection_started_date",
                y="inspection_worktime_minute/input_data_count",
                title="画像あたり検査時間の折れ線グラフ",
                x_axis_label="1回目の検査開始日",
                y_axis_label="画像あたり検査時間[min]",
            ),
        ]
        write_graph(fig_info_list_input_data_count, html_title="折れ線-横軸_検査開始日-縦軸_画像あたりの指標-検査者用")

    def write_productivity_line_graph_for_acceptor(
        self,
        df: pandas.DataFrame,
        first_acceptance_user_id_list: Optional[List[str]] = None,
    ):
        """
        日毎の生産性を受入作業者ごとにプロットする。

        Args:
            df:
            first_acceptance_user_id_list:

        Returns:

        """

        def write_graph(fig_info_list: List[Dict[str, str]], html_title: str):
            """

            Args:
                fig_info_list:
                html_title:

            Returns:

            """
            output_file = f"{self.line_graph_outdir}/{html_title}.html"
            logger.debug(f"{output_file} を出力します。")

            figs: List[bokeh.plotting.Figure] = []
            for fig_info in fig_info_list:
                figs.append(
                    figure(
                        plot_width=1200,
                        plot_height=600,
                        title=fig_info["title"],
                        x_axis_label=fig_info["x_axis_label"],
                        x_axis_type="datetime",
                        y_axis_label=fig_info["y_axis_label"],
                    )
                )

            for user_index, user_id in enumerate(user_id_list):
                filtered_df = df[df["first_acceptance_user_id"] == user_id]
                if filtered_df.empty:
                    logger.debug(f"dataframe is empty. user_id = {user_id}")
                    continue

                source = ColumnDataSource(data=filtered_df)
                color = self.my_palette[user_index]
                username = filtered_df.iloc[0]["first_acceptance_username"]

                for fig, fig_info in zip(figs, fig_info_list):
                    self._plot_line_and_circle(
                        fig,
                        x_column_name=fig_info["x"],
                        y_column_name=fig_info["y"],
                        source=source,
                        legend_label=username,
                        color=color,
                    )

            hover_tool = self._create_hover_tool(tooltip_item)
            for fig in figs:
                self._set_legend(fig, hover_tool)

            bokeh.plotting.reset_output()
            bokeh.plotting.output_file(output_file, title=html_title)
            bokeh.plotting.save(bokeh.layouts.column(figs))

        tooltip_item = [
            "first_acceptance_user_id",
            "first_acceptance_username",
            "first_acceptance_started_date",
            "first_acceptance_worktime_hour",
            "acceptance_worktime_hour",
            "annotation_count",
            "input_data_count",
            "task_count",
            "inspection_count",
        ]

        if len(df) == 0:
            logger.info("データが0件のため出力しない")
            return

        user_id_list = self.create_user_id_list(
            df,
            "first_acceptance_user_id",
            datetime_column="first_acceptance_started_date",
            arg_user_id_list=first_acceptance_user_id_list,
        )
        logger.debug(f"受入者用の折れ線グラフに表示する、受入者のuser_id = {user_id_list}")

        df["date_first_acceptance_started_date"] = df["first_acceptance_started_date"].map(
            lambda e: dateutil.parser.parse(e).date()
        )

        fig_info_list_acceptance_count = [
            dict(
                x="date_first_acceptance_started_date",
                y="first_acceptance_worktime_minute/annotation_count",
                title="アノテーションあたり受入時間(1回目)の折れ線グラフ",
                x_axis_label="1回目の受入開始日",
                y_axis_label="アノテーションあたり受入時間(1回目)[min]",
            ),
            dict(
                x="date_first_acceptance_started_date",
                y="acceptance_worktime_minute/annotation_count",
                title="アノテーションあたり受入時間の折れ線グラフ",
                x_axis_label="1回目の受入開始日",
                y_axis_label="アノテーションあたり受入時間[min]",
            ),
        ]
        write_graph(
            fig_info_list_acceptance_count,
            html_title="折れ線-横軸_受入開始日-縦軸_アノテーションあたりの指標-受入者用",
        )

        fig_info_list_input_data_count = [
            dict(
                x="date_first_acceptance_started_date",
                y="first_acceptance_worktime_minute/input_data_count",
                title="画像あたり受入時間(1回目)の折れ線グラフ",
                x_axis_label="1回目の受入開始日",
                y_axis_label="画像あたり受入時間(1回目)[min]",
            ),
            dict(
                x="date_first_acceptance_started_date",
                y="acceptance_worktime_minute/input_data_count",
                title="画像あたり受入時間の折れ線グラフ",
                x_axis_label="1回目の受入開始日",
                y_axis_label="画像あたり受入時間[min]",
            ),
        ]
        write_graph(fig_info_list_input_data_count, html_title="折れ線-横軸_受入開始日-縦軸_画像あたりの指標-受入者用")

    def write_cumulative_line_graph_for_annotator(
        self,
        df: pandas.DataFrame,
        output_target_list: Optional[List[OutputTarget]] = None,
        first_annotation_user_id_list: Optional[List[str]] = None,
    ):
        """
        教師付作業者用の累積折れ線グラフを出力する。

        Args:
            df:
            output_target_list: すべて出力する
            first_annotation_user_id_list: 最初のアノテーションを担当したuser_idのList. Noneの場合はtask_dfから決まる。

        Returns:

        """

        def write_cumulative_graph(fig_info_list: List[Dict[str, str]], html_title: str):
            """
            累計グラフを出力する。

            Args:
                fig_info_list:
                html_title:

            Returns:

            """
            output_file = f"{self.line_graph_outdir}/{html_title}.html"
            logger.debug(f"{output_file} を出力します。")

            figs: List[bokeh.plotting.Figure] = []
            for fig_info in fig_info_list:
                figs.append(
                    figure(
                        plot_width=1200,
                        plot_height=600,
                        title=fig_info["title"],
                        x_axis_label=fig_info["x_axis_label"],
                        y_axis_label=fig_info["y_axis_label"],
                    )
                )

            for user_index, user_id in enumerate(user_id_list):
                filtered_df = df[df["first_annotation_user_id"] == user_id]
                if filtered_df.empty:
                    logger.debug(f"dataframe is empty. user_id = {user_id}")
                    continue

                source = ColumnDataSource(data=filtered_df)
                color = self.my_palette[user_index]
                username = filtered_df.iloc[0]["first_annotation_username"]

                for fig, fig_info in zip(figs, fig_info_list):
                    self._plot_line_and_circle(
                        fig,
                        x_column_name=fig_info["x"],
                        y_column_name=fig_info["y"],
                        source=source,
                        legend_label=username,
                        color=color,
                    )

            hover_tool = self._create_hover_tool(tooltip_item)
            for fig in figs:
                self._set_legend(fig, hover_tool)

            bokeh.plotting.reset_output()
            bokeh.plotting.output_file(output_file, title=html_title)
            bokeh.plotting.save(bokeh.layouts.column(figs))

        tooltip_item = [
            "task_id",
            "phase",
            "status",
            "first_annotation_user_id",
            "first_annotation_username",
            "first_annotation_started_datetime",
            "updated_datetime",
            "annotation_worktime_hour",
            "inspection_worktime_hour",
            "acceptance_worktime_hour",
            "annotation_count",
            "input_data_count",
            "inspection_count",
        ]

        if len(df) == 0:
            logger.info("タスク一覧が0件のため出力しない")
            return

        user_id_list = self.create_user_id_list(
            df,
            "first_annotation_user_id",
            datetime_column="first_annotation_started_datetime",
            arg_user_id_list=first_annotation_user_id_list,
        )
        logger.debug(f"教師付者用の累積折れ線グラフに表示する、教師付者のuser_id = {user_id_list}")

        if output_target_list is None or OutputTarget.ANNOTATION in output_target_list:
            # 横軸が累計のアノテーション数
            fig_info_list_annotation_count = [
                dict(
                    x="cumulative_annotation_count",
                    y="cumulative_annotation_worktime_hour",
                    title="アノテーション数と教師付時間の累積グラフ",
                    x_axis_label="アノテーション数",
                    y_axis_label="教師付時間[hour]",
                ),
                dict(
                    x="cumulative_annotation_count",
                    y="cumulative_first_annotation_worktime_hour",
                    title="アノテーション数と1回目の教師付時間の累積グラフ",
                    x_axis_label="アノテーション数",
                    y_axis_label="1回目の教師付時間[hour]",
                ),
                dict(
                    x="cumulative_annotation_count",
                    y="cumulative_inspection_worktime_hour",
                    title="アノテーション数と検査時間の累積グラフ",
                    x_axis_label="アノテーション数",
                    y_axis_label="検査時間[hour]",
                ),
                dict(
                    x="cumulative_annotation_count",
                    y="cumulative_acceptance_worktime_hour",
                    title="アノテーション数と受入時間の累積グラフ",
                    x_axis_label="アノテーション数",
                    y_axis_label="受入時間[hour]",
                ),
                dict(
                    x="cumulative_annotation_count",
                    y="cumulative_inspection_count",
                    title="アノテーション数と検査コメント数の累積グラフ",
                    x_axis_label="アノテーション数",
                    y_axis_label="検査コメント数",
                ),
            ]
            write_cumulative_graph(fig_info_list_annotation_count, html_title="累積折れ線-横軸_アノテーション数-教師付者用")

        if output_target_list is None or OutputTarget.INPUT_DATA in output_target_list:
            # 横軸が累計の入力データ数
            fig_info_list_input_data_count = [
                dict(
                    x="cumulative_input_data_count",
                    y="cumulative_annotation_worktime_hour",
                    title="入力データ数と教師付時間の累積グラフ",
                    x_axis_label="入力データ数",
                    y_axis_label="教師付時間[hour]",
                ),
                dict(
                    x="cumulative_input_data_count",
                    y="cumulative_inspection_worktime_hour",
                    title="入力データ数と検査時間の累積グラフ",
                    x_axis_label="入力データ数",
                    y_axis_label="検査時間[hour]",
                ),
                dict(
                    x="cumulative_input_data_count",
                    y="cumulative_acceptance_worktime_hour",
                    title="入力データ数と受入時間の累積グラフ",
                    x_axis_label="入力データ数",
                    y_axis_label="受入時間[hour]",
                ),
                dict(
                    x="cumulative_input_data_count",
                    y="cumulative_inspection_count",
                    title="アノテーション数と検査コメント数の累積グラフ",
                    x_axis_label="入力データ数",
                    y_axis_label="検査コメント数",
                ),
            ]
            write_cumulative_graph(fig_info_list_input_data_count, html_title="累積折れ線-横軸_入力データ数-教師付者用")

        if output_target_list is None or OutputTarget.TASK in output_target_list:
            # 横軸が累計のタスク数
            fig_info_list_task_count = [
                dict(
                    x="cumulative_task_count",
                    y="cumulative_annotation_worktime_hour",
                    title="タスク数と教師付時間の累積グラフ",
                    x_axis_label="タスク数",
                    y_axis_label="教師付時間[hour]",
                ),
                dict(
                    x="cumulative_task_count",
                    y="cumulative_inspection_worktime_hour",
                    title="タスク数と検査時間の累積グラフ",
                    x_axis_label="タスク数",
                    y_axis_label="検査時間[hour]",
                ),
                dict(
                    x="cumulative_task_count",
                    y="cumulative_acceptance_worktime_hour",
                    title="タスク数と受入時間の累積グラフ",
                    x_axis_label="タスク数",
                    y_axis_label="受入時間[hour]",
                ),
                dict(
                    x="cumulative_task_count",
                    y="cumulative_inspection_count",
                    title="タスク数と検査コメント数の累積グラフ",
                    x_axis_label="タスク数",
                    y_axis_label="検査コメント数",
                ),
                dict(
                    x="cumulative_task_count",
                    y="cumulative_number_of_rejections",
                    title="タスク数と差し戻し回数の累積グラフ",
                    x_axis_label="タスク数",
                    y_axis_label="差し戻し回数",
                ),
                dict(
                    x="cumulative_task_count",
                    y="cumulative_number_of_rejections_by_inspection",
                    title="タスク数と差し戻し回数(検査フェーズ)の累積グラフ",
                    x_axis_label="タスク数",
                    y_axis_label="差し戻し回数(検査フェーズ)",
                ),
                dict(
                    x="cumulative_task_count",
                    y="cumulative_number_of_rejections_by_acceptance",
                    title="タスク数と差し戻し回数(受入フェーズ)の累積グラフ",
                    x_axis_label="タスク数",
                    y_axis_label="差し戻し回数(受入フェーズ)",
                ),
            ]
            write_cumulative_graph(fig_info_list_task_count, html_title="累積折れ線-横軸_タスク数-教師付者用")

    def write_cumulative_line_graph_for_inspector(
        self,
        df: pandas.DataFrame,
        output_target_list: Optional[List[OutputTarget]] = None,
        first_inspection_user_id_list: Optional[List[str]] = None,
    ):
        """
        検査作業者用の累積折れ線グラフを出力する。

        Args:
            df:
            first_inspection_user_id_list: 最初の検査フェーズを担当したuser_idのList. Noneの場合はtask_dfから決まる。

        Returns:

        """

        def write_cumulative_graph(fig_info_list: List[Dict[str, str]], html_title: str):
            """
            累計グラフを出力する。

            Args:
                fig_info_list:
                html_title:

            Returns:

            """
            output_file = f"{self.line_graph_outdir}/{html_title}.html"
            logger.debug(f"{output_file} を出力します。")

            figs: List[bokeh.plotting.Figure] = []
            for fig_info in fig_info_list:
                figs.append(
                    figure(
                        plot_width=1200,
                        plot_height=600,
                        title=fig_info["title"],
                        x_axis_label=fig_info["x_axis_label"],
                        y_axis_label=fig_info["y_axis_label"],
                    )
                )

            for user_index, user_id in enumerate(user_id_list):
                filtered_df = df[df["first_inspection_user_id"] == user_id]
                if filtered_df.empty:
                    logger.debug(f"dataframe is empty. user_id = {user_id}")
                    continue

                source = ColumnDataSource(data=filtered_df)
                color = self.my_palette[user_index]
                username = filtered_df.iloc[0]["first_inspection_username"]

                for fig, fig_info in zip(figs, fig_info_list):
                    self._plot_line_and_circle(
                        fig,
                        x_column_name=fig_info["x"],
                        y_column_name=fig_info["y"],
                        source=source,
                        legend_label=username,
                        color=color,
                    )

            hover_tool = self._create_hover_tool(tooltip_item)
            for fig in figs:
                self._set_legend(fig, hover_tool)

            bokeh.plotting.reset_output()
            bokeh.plotting.output_file(output_file, title=html_title)
            bokeh.plotting.save(bokeh.layouts.column(figs))

        tooltip_item = [
            "task_id",
            "phase",
            "status",
            "first_annotation_username",
            "first_annotation_started_datetime",
            "first_inspection_username",
            "first_inspection_started_datetime",
            "updated_datetime",
            "annotation_worktime_hour",
            "inspection_worktime_hour",
            "acceptance_worktime_hour",
            "annotation_count",
            "input_data_count",
            "inspection_count",
        ]

        if len(df) == 0:
            logger.info("タスク一覧が0件のため出力しない")
            return

        user_id_list = self.create_user_id_list(
            df,
            "first_inspection_user_id",
            datetime_column="first_inspection_started_datetime",
            arg_user_id_list=first_inspection_user_id_list,
        )

        if len(user_id_list) == 0:
            logger.info(f"検査フェーズを担当してユーザがいないため、検査者用のグラフは出力しません。")
            return

        logger.debug(f"検査者用の累積折れ線グラフに表示する、検査者のuser_id = {user_id_list}")

        if output_target_list is None or OutputTarget.ANNOTATION in output_target_list:
            # 横軸が累計のアノテーション数
            fig_info_list_annotation_count = [
                dict(
                    x="cumulative_annotation_count",
                    y="cumulative_first_inspection_worktime_hour",
                    title="アノテーション数と1回目検査時間の累積グラフ",
                    x_axis_label="アノテーション数",
                    y_axis_label="1回目の検査時間[hour]",
                ),
                dict(
                    x="cumulative_annotation_count",
                    y="cumulative_inspection_worktime_hour",
                    title="アノテーション数と検査時間の累積グラフ",
                    x_axis_label="アノテーション数",
                    y_axis_label="検査時間[hour]",
                ),
                dict(
                    x="cumulative_inspection_count",
                    y="cumulative_first_inspection_worktime_hour",
                    title="検査コメント数と1回目検査時間の累積グラフ",
                    x_axis_label="検査コメント数",
                    y_axis_label="1回目の検査時間[hour]",
                ),
                dict(
                    x="cumulative_inspection_count",
                    y="cumulative_inspection_worktime_hour",
                    title="検査コメント数と検査時間の累積グラフ",
                    x_axis_label="検査コメント数",
                    y_axis_label="検査時間[hour]",
                ),
            ]
            write_cumulative_graph(fig_info_list_annotation_count, html_title="累積折れ線-横軸_アノテーション数-検査者用")

        if output_target_list is None or OutputTarget.INPUT_DATA in output_target_list:
            # 横軸が累計の入力データ数
            fig_info_list_input_data_count = [
                dict(
                    x="cumulative_input_data_count",
                    y="cumulative_first_inspection_worktime_hour",
                    title="入力データ数と1回目検査時間の累積グラフ",
                    x_axis_label="入力データ数",
                    y_axis_label="1回目の検査時間[hour]",
                ),
                dict(
                    x="cumulative_input_data_count",
                    y="cumulative_inspection_worktime_hour",
                    title="入力データ数と検査時間の累積グラフ",
                    x_axis_label="入力データ数",
                    y_axis_label="検査時間[hour]",
                ),
            ]
            write_cumulative_graph(fig_info_list_input_data_count, html_title="累積折れ線-横軸_入力データ数-検査者用")

    def write_cumulative_line_graph_for_acceptor(
        self,
        df: pandas.DataFrame,
        output_target_list: Optional[List[OutputTarget]] = None,
        first_acceptance_user_id_list: Optional[List[str]] = None,
    ):
        """
        受入者用の累積折れ線グラフを出力する。

        Args:
            df:
            first_acceptance_user_id_list: 最初の検査フェーズを担当したuser_idのList. Noneの場合はtask_dfから決まる。

        Returns:

        """

        def write_cumulative_graph(fig_info_list: List[Dict[str, str]], html_title: str):
            """
            累計グラフを出力する。

            Args:
                fig_info_list:
                html_title:

            Returns:

            """
            output_file = f"{self.line_graph_outdir}/{html_title}.html"
            logger.debug(f"{output_file} を出力します。")

            figs: List[bokeh.plotting.Figure] = []
            for fig_info in fig_info_list:
                figs.append(
                    figure(
                        plot_width=1200,
                        plot_height=600,
                        title=fig_info["title"],
                        x_axis_label=fig_info["x_axis_label"],
                        y_axis_label=fig_info["y_axis_label"],
                    )
                )

            for user_index, user_id in enumerate(user_id_list):
                filtered_df = df[df["first_acceptance_user_id"] == user_id]
                if filtered_df.empty:
                    logger.debug(f"dataframe is empty. user_id = {user_id}")
                    continue

                source = ColumnDataSource(data=filtered_df)
                color = self.my_palette[user_index]
                username = filtered_df.iloc[0]["first_acceptance_username"]

                for fig, fig_info in zip(figs, fig_info_list):
                    self._plot_line_and_circle(
                        fig,
                        x_column_name=fig_info["x"],
                        y_column_name=fig_info["y"],
                        source=source,
                        legend_label=username,
                        color=color,
                    )

            hover_tool = self._create_hover_tool(tooltip_item)
            for fig in figs:
                self._set_legend(fig, hover_tool)

            bokeh.plotting.reset_output()
            bokeh.plotting.output_file(output_file, title=html_title)
            bokeh.plotting.save(bokeh.layouts.column(figs))

        tooltip_item = [
            "task_id",
            "phase",
            "status",
            "first_annotation_username",
            "first_annotation_started_datetime",
            "first_inspection_username",
            "first_inspection_started_datetime",
            "first_acceptance_username",
            "first_acceptance_started_datetime",
            "updated_datetime",
            "annotation_worktime_hour",
            "inspection_worktime_hour",
            "acceptance_worktime_hour",
            "annotation_count",
            "input_data_count",
            "inspection_count",
        ]

        if len(df) == 0:
            logger.info("タスク一覧が0件のため出力しない")
            return

        user_id_list = self.create_user_id_list(
            df,
            "first_acceptance_user_id",
            datetime_column="first_acceptance_started_datetime",
            arg_user_id_list=first_acceptance_user_id_list,
        )

        if len(user_id_list) == 0:
            logger.info(f"受入フェーズを担当してユーザがいないため、受入者用のグラフは出力しません。")
            return

        logger.debug(f"受入者用の累積折れ線グラフに表示する、受入者のuser_id = {user_id_list}")

        if output_target_list is None or OutputTarget.ANNOTATION in output_target_list:
            # 横軸が累計のアノテーション数
            fig_info_list_annotation_count = [
                dict(
                    x="cumulative_annotation_count",
                    y="cumulative_first_acceptance_worktime_hour",
                    title="アノテーション数と1回目受入時間の累積グラフ",
                    x_axis_label="アノテーション数",
                    y_axis_label="1回目の受入時間[hour]",
                ),
                dict(
                    x="cumulative_annotation_count",
                    y="cumulative_acceptance_worktime_hour",
                    title="アノテーション数と受入時間の累積グラフ",
                    x_axis_label="アノテーション数",
                    y_axis_label="受入時間[hour]",
                ),
                dict(
                    x="cumulative_inspection_count",
                    y="cumulative_first_acceptance_worktime_hour",
                    title="検査コメント数と1回目受入時間の累積グラフ",
                    x_axis_label="検査コメント数",
                    y_axis_label="1回目の受入時間[hour]",
                ),
                dict(
                    x="cumulative_inspection_count",
                    y="cumulative_acceptance_worktime_hour",
                    title="検査コメント数と受入時間の累積グラフ",
                    x_axis_label="検査コメント数",
                    y_axis_label="受入時間[hour]",
                ),
            ]
            write_cumulative_graph(fig_info_list_annotation_count, html_title="累積折れ線-横軸_アノテーション数-受入者用")

        if output_target_list is None or OutputTarget.INPUT_DATA in output_target_list:
            # 横軸が累計の入力データ数
            fig_info_list_input_data_count = [
                dict(
                    x="cumulative_input_data_count",
                    y="cumulative_first_acceptance_worktime_hour",
                    title="入力データ数と1回目受入時間の累積グラフ",
                    x_axis_label="入力データ数",
                    y_axis_label="1回目の受入時間[hour]",
                ),
                dict(
                    x="cumulative_input_data_count",
                    y="cumulative_acceptance_worktime_hour",
                    title="入力データ数と受入時間の累積グラフ",
                    x_axis_label="入力データ数",
                    y_axis_label="受入時間[hour]",
                ),
            ]
            write_cumulative_graph(fig_info_list_input_data_count, html_title="累積折れ線-横軸_入力データ数-受入者用")

    def write_cumulative_line_graph_by_date(
        self, df: pandas.DataFrame, user_id_list: Optional[List[str]] = None
    ) -> None:
        """
        ユーザごとの作業時間を、日単位でプロットする。

        Args:
            user_id_list: プロットするユーザのuser_id


        """
        tooltip_item = [
            "user_id",
            "username",
            "date",
            "worktime_hour",
        ]

        if len(df) == 0:
            logger.info("データが0件のため出力ません。")
            return

        html_title = "累積折れ線-横軸_日-縦軸_作業時間"
        output_file = f"{self.line_graph_outdir}/{html_title}.html"

        user_id_list = self.create_user_id_list(df, "user_id", datetime_column="date", arg_user_id_list=user_id_list)
        if len(user_id_list) == 0:
            logger.info(f"作業しているユーザがいないため、{html_title} グラフは出力しません。")
            return

        logger.debug(f"{html_title} グラフに表示するユーザのuser_id = {user_id_list}")

        # 日付変換
        df["date_date"] = df["date"].map(lambda e: dateutil.parser.parse(e).date())

        fig_info_list = [
            dict(
                x="date_date",
                y="cumulative_worktime_hour",
                title="累積作業時間",
                x_axis_label="日",
                y_axis_label="作業時間[hour]",
            )
        ]

        logger.debug(f"{output_file} を出力します。")

        figs: List[bokeh.plotting.Figure] = []
        for fig_info in fig_info_list:
            figs.append(
                figure(
                    plot_width=1200,
                    plot_height=600,
                    title=fig_info["title"],
                    x_axis_label=fig_info["x_axis_label"],
                    x_axis_type="datetime",
                    y_axis_label=fig_info["y_axis_label"],
                )
            )

        for user_index, user_id in enumerate(user_id_list):
            filtered_df = df[df["user_id"] == user_id]
            if filtered_df.empty:
                logger.debug(f"dataframe is empty. user_id = {user_id}")
                continue

            source = ColumnDataSource(data=filtered_df)
            color = self.my_palette[user_index]
            username = filtered_df.iloc[0]["username"]

            for fig, fig_info in zip(figs, fig_info_list):
                self._plot_line_and_circle(
                    fig,
                    x_column_name=fig_info["x"],
                    y_column_name=fig_info["y"],
                    source=source,
                    legend_label=username,
                    color=color,
                )

        hover_tool = self._create_hover_tool(tooltip_item)
        for fig in figs:
            self._set_legend(fig, hover_tool)

        bokeh.plotting.reset_output()
        bokeh.plotting.output_file(output_file, title=html_title)
        bokeh.plotting.save(bokeh.layouts.column(figs))

    def write_cumulative_line_graph_overall(self, df: pandas.DataFrame) -> None:
        """
        プロジェクト全体に対して、累積作業時間の折れ線グラフを出力する。

        Args:
            df: タスク一覧のDataFrame
        """

        tooltip_item = [
            "task_id",
            "phase",
            "status",
            "first_annotation_started_datetime",
            "updated_datetime",
            "annotation_worktime_hour",
            "inspection_worktime_hour",
            "acceptance_worktime_hour",
            "sum_worktime_hour",
            "annotation_count",
            "input_data_count",
            "inspection_count",
        ]
        if len(df) == 0:
            logger.info("データが0件のため出力ません。")
            return

        html_title = "累積折れ線-横軸_アノテーション数-縦軸_作業時間-全体"
        output_file = f"{self.line_graph_outdir}/{html_title}.html"

        logger.debug(f"{output_file} を出力します。")

        fig = figure(
            plot_width=1200,
            plot_height=600,
            title="アノテーション数と作業時間の累積グラフ",
            x_axis_label="アノテーション数",
            y_axis_label="作業時間[hour]",
        )

        fig_info_list = [
            dict(x="cumulative_annotation_count", y="cumulative_sum_worktime_hour", legend_label="sum"),
            dict(x="cumulative_annotation_count", y="cumulative_annotation_worktime_hour", legend_label="annotation"),
            dict(x="cumulative_annotation_count", y="cumulative_inspection_worktime_hour", legend_label="inspection"),
            dict(x="cumulative_annotation_count", y="cumulative_acceptance_worktime_hour", legend_label="acceptance"),
        ]

        source = ColumnDataSource(data=df)

        for index, fig_info in enumerate(fig_info_list):
            color = self.my_palette[index]

            self._plot_line_and_circle(
                fig,
                x_column_name=fig_info["x"],
                y_column_name=fig_info["y"],
                source=source,
                legend_label=fig_info["legend_label"],
                color=color,
            )
        hover_tool = self._create_hover_tool(tooltip_item)
        self._set_legend(fig, hover_tool)

        bokeh.plotting.reset_output()
        bokeh.plotting.output_file(output_file, title=html_title)
        bokeh.plotting.save(bokeh.layouts.column([fig]))

    def write_whole_productivity_line_graph(self, df: pandas.DataFrame):
        """
        全体の生産量や生産性をプロットする


        Args:
            df:

        Returns:

        """

        def create_figure(title: str, y_axis_label: str) -> bokeh.plotting.Figure:
            return figure(
                plot_width=1200,
                plot_height=600,
                title=title,
                x_axis_label="日",
                x_axis_type="datetime",
                y_axis_label=y_axis_label,
            )

        def plot_and_moving_average(fig, y_column_name: str, legend_name: str, source, color, **kwargs):
            x_column_name = "dt_date"

            # 値をプロット
            self._plot_line_and_circle(
                fig,
                x_column_name=x_column_name,
                y_column_name=y_column_name,
                source=source,
                color=color,
                legend_label=legend_name,
                **kwargs,
            )

            # 移動平均をプロット
            self._plot_moving_average(
                fig,
                x_column_name=x_column_name,
                y_column_name=f"{y_column_name}__lastweek",
                source=source,
                color=color,
                legend_label=f"{legend_name}の1週間移動平均",
                **kwargs,
            )

        def create_task_figure():
            y_range_name = "worktime_axis"
            fig_task = create_figure(title="日ごとのタスク数と作業時間", y_axis_label="タスク数")
            fig_task.add_layout(
                LinearAxis(
                    y_range_name=y_range_name,
                    axis_label="作業時間[hour]",
                ),
                "right",
            )
            y_overlimit = 0.05
            fig_task.extra_y_ranges = {
                y_range_name: DataRange1d(end=df["actual_worktime_hour"].max() * (1 + y_overlimit))
            }
            plot_and_moving_average(
                fig=fig_task,
                y_column_name="task_count",
                legend_name="タスク数",
                source=source,
                color=self.my_small_palette[0],
            )
            plot_and_moving_average(
                fig=fig_task,
                y_column_name="actual_worktime_hour",
                legend_name="実績作業時間",
                source=source,
                color=self.my_small_palette[1],
                y_range_name=y_range_name,
            )
            return fig_task

        def create_input_data_figure():
            y_range_name = "worktime_axis"
            fig_input_data = create_figure(title="日ごとの入力データ数と作業時間", y_axis_label="入力データ数")
            fig_input_data.add_layout(
                LinearAxis(
                    y_range_name=y_range_name,
                    axis_label="作業時間[hour]",
                ),
                "right",
            )
            y_overlimit = 0.05
            fig_input_data.extra_y_ranges = {
                y_range_name: DataRange1d(end=df["actual_worktime_hour"].max() * (1 + y_overlimit))
            }
            plot_and_moving_average(
                fig=fig_input_data,
                y_column_name="input_data_count",
                legend_name="入力データ数",
                source=source,
                color=self.my_small_palette[0],
            )
            plot_and_moving_average(
                fig=fig_input_data,
                y_column_name="actual_worktime_hour",
                legend_name="実績作業時間",
                source=source,
                color=self.my_small_palette[1],
                y_range_name=y_range_name,
            )
            return fig_input_data

        if len(df) == 0:
            logger.info("データが0件のため出力しない")
            return

        df["dt_date"] = df["date"].map(lambda e: dateutil.parser.parse(e).date())

        html_title = "折れ線-横軸_日-全体"
        output_file = f"{self.line_graph_outdir}/{html_title}.html"
        logger.debug(f"{output_file} を出力します。")

        fig_list = [
            create_figure(title="日ごとの作業時間", y_axis_label="作業時間[hour]"),
            create_figure(title="日ごとのタスクあたり作業時間", y_axis_label="タスクあたり作業時間[hour/task]"),
            create_figure(title="日ごとの入力データあたり作業時間", y_axis_label="入力データあたり作業時間[hour/input_data]"),
            create_figure(title="日ごとのアノテーションあたり作業時間", y_axis_label="アノテーションあたり作業時間[hour/annotation]"),
        ]

        fig_info_list = [
            {
                "x": "dt_date",
                "y_info_list": [
                    {"column": "actual_worktime_hour", "legend": "実績作業時間"},
                    {"column": "monitored_worktime_hour", "legend": "計測作業時間"},
                ],
            },
            {
                "x": "dt_date",
                "y_info_list": [
                    {"column": "actual_worktime_hour/task_count", "legend": "タスクあたり実績作業時間"},
                    {"column": "monitored_worktime_hour/task_count", "legend": "タスクあたり計測作業時間"},
                ],
            },
            {
                "x": "dt_date",
                "y_info_list": [
                    {"column": "actual_worktime_hour/input_data_count", "legend": "入力データあたり実績作業時間"},
                    {"column": "monitored_worktime_hour/input_data_count", "legend": "入力データあたり計測作業時間"},
                ],
            },
            {
                "x": "dt_date",
                "y_info_list": [
                    {"column": "actual_worktime_hour/annotation_count", "legend": "アノテーションあたり実績作業時間"},
                    {"column": "monitored_worktime_hour/annotation_count", "legend": "アノテーションあたり計測作業時間"},
                ],
            },
        ]

        MOVING_WINDOW_SIZE = 7
        for column in ["task_count", "input_data_count", "actual_worktime_hour", "monitored_worktime_hour"]:
            df[f"{column}__lastweek"] = df[column].rolling(MOVING_WINDOW_SIZE).mean()

        source = ColumnDataSource(data=df)

        for fig, fig_info in zip(fig_list, fig_info_list):
            y_info_list: List[Dict[str, str]] = fig_info["y_info_list"]  # type: ignore
            for index, y_info in enumerate(y_info_list):
                color = self.my_small_palette[index]

                plot_and_moving_average(
                    fig=fig, y_column_name=y_info["column"], legend_name=y_info["legend"], source=source, color=color
                )

        tooltip_item = [
            "date",
            "task_count",
            "input_data_count",
            "actual_worktime_hour",
            "monitored_worktime_hour",
            "cumsum_task_count",
            "cumsum_input_data_count",
            "cumsum_actual_worktime_hour",
            "actual_worktime_hour/task_count",
            "actual_worktime_hour/input_data_count",
            "actual_worktime_hour/annotation_count",
            "monitored_worktime_hour/task_count",
            "monitored_worktime_hour/input_data_count",
            "monitored_worktime_hour/annotation_count",
        ]
        hover_tool = self._create_hover_tool(tooltip_item)

        fig_list.insert(0, create_task_figure())
        fig_list.insert(1, create_input_data_figure())

        for fig in fig_list:
            self._set_legend(fig, hover_tool)

        bokeh.plotting.reset_output()
        bokeh.plotting.output_file(output_file, title=html_title)
        bokeh.plotting.save(bokeh.layouts.column(fig_list))

    def write_whole_cumulative_line_graph(self, df: pandas.DataFrame):
        """
        全体の生産量や作業時間の累積折れ線グラフを出力する
        """

        def create_figure(title: str, y_axis_label: str) -> bokeh.plotting.Figure:
            return figure(
                plot_width=1200,
                plot_height=600,
                title=title,
                x_axis_label="日",
                x_axis_type="datetime",
                y_axis_label=y_axis_label,
            )

        def create_task_figure():
            y_range_name = "worktime_axis"
            fig = create_figure(title="日ごとの累積タスク数と累積作業時間", y_axis_label="タスク数")
            fig.add_layout(
                LinearAxis(
                    y_range_name=y_range_name,
                    axis_label="作業時間[hour]",
                ),
                "right",
            )
            y_overlimit = 0.05
            fig.extra_y_ranges = {
                y_range_name: DataRange1d(end=df["cumsum_actual_worktime_hour"].max() * (1 + y_overlimit))
            }

            # 値をプロット
            self._plot_line_and_circle(
                fig,
                x_column_name=x_column_name,
                y_column_name="cumsum_task_count",
                source=source,
                color=self.my_small_palette[0],
                legend_label="タスク数",
            )

            # 値をプロット
            self._plot_line_and_circle(
                fig,
                x_column_name=x_column_name,
                y_column_name="cumsum_actual_worktime_hour",
                source=source,
                color=self.my_small_palette[1],
                legend_label="実績作業時間",
                y_range_name=y_range_name,
            )

            return fig

        def create_input_data_figure():
            y_range_name = "worktime_axis"
            fig = create_figure(title="日ごとの累積入力データ数と累積作業時間", y_axis_label="入力データ数")
            fig.add_layout(
                LinearAxis(
                    y_range_name=y_range_name,
                    axis_label="作業時間[hour]",
                ),
                "right",
            )
            y_overlimit = 0.05
            fig.extra_y_ranges = {
                y_range_name: DataRange1d(end=df["cumsum_actual_worktime_hour"].max() * (1 + y_overlimit))
            }

            # 値をプロット
            self._plot_line_and_circle(
                fig,
                x_column_name=x_column_name,
                y_column_name="cumsum_input_data_count",
                source=source,
                color=self.my_small_palette[0],
                legend_label="入力データ数",
            )

            # 値をプロット
            self._plot_line_and_circle(
                fig,
                x_column_name=x_column_name,
                y_column_name="cumsum_actual_worktime_hour",
                source=source,
                color=self.my_small_palette[1],
                legend_label="実績作業時間",
                y_range_name=y_range_name,
            )

            return fig

        if len(df) == 0:
            logger.info("データが0件のため出力しない")
            return

        df["dt_date"] = df["date"].map(lambda e: dateutil.parser.parse(e).date())
        df["cumsum_monitored_worktime_hour"] = df["monitored_worktime_hour"].cumsum()

        html_title = "累積折れ線-横軸_日-全体"
        output_file = f"{self.line_graph_outdir}/{html_title}.html"
        logger.debug(f"{output_file} を出力します。")

        fig_list = [
            create_figure(title="日ごとの累積作業時間", y_axis_label="作業時間[hour]"),
        ]

        fig_info_list = [
            {
                "x": "dt_date",
                "y_info_list": [
                    {"column": "cumsum_actual_worktime_hour", "legend": "実績作業時間"},
                    {"column": "cumsum_monitored_worktime_hour", "legend": "計測作業時間"},
                ],
            },
        ]

        source = ColumnDataSource(data=df)

        for fig, fig_info in zip(fig_list, fig_info_list):
            x_column_name = "dt_date"
            y_info_list: List[Dict[str, str]] = fig_info["y_info_list"]  # type: ignore
            for index, y_info in enumerate(y_info_list):
                color = self.my_small_palette[index]

                # 値をプロット
                self._plot_line_and_circle(
                    fig,
                    x_column_name=x_column_name,
                    y_column_name=y_info["column"],
                    source=source,
                    color=color,
                    legend_label=y_info["legend"],
                )

        tooltip_item = [
            "date",
            "task_count",
            "input_data_count",
            "actual_worktime_hour",
            "monitored_worktime_hour",
            "cumsum_task_count",
            "cumsum_input_data_count",
            "cumsum_actual_worktime_hour",
            "cumsum_monitored_worktime_hour",
        ]
        hover_tool = self._create_hover_tool(tooltip_item)

        fig_list.insert(0, create_task_figure())
        fig_list.insert(1, create_input_data_figure())

        for fig in fig_list:
            self._set_legend(fig, hover_tool)

        bokeh.plotting.reset_output()
        bokeh.plotting.output_file(output_file, title=html_title)
        bokeh.plotting.save(bokeh.layouts.column(fig_list))

    def write_gradient_graph_for_annotator(
        self,
        df: pandas.DataFrame,
        first_annotation_user_id_list: Optional[List[str]] = None,
    ):
        """
        教師付作業者用の累積折れ線グラフの傾きを出力する。

        Args:
            df:
            first_annotation_user_id_list: 最初のアノテーションを担当したuser_idのList. Noneの場合はtask_dfから決まる。

        Returns:

        """

        def write_cumulative_graph(fig_info_list: List[Dict[str, str]], html_title: str):
            """
            累計グラフを出力する。

            Args:
                fig_info_list:
                html_title:

            Returns:

            """
            output_file = f"{self.line_graph_outdir}/{html_title}.html"
            logger.debug(f"{output_file} を出力します。")

            figs: List[bokeh.plotting.Figure] = []
            for fig_info in fig_info_list:
                figs.append(
                    figure(
                        plot_width=1200,
                        plot_height=600,
                        title=fig_info["title"],
                        x_axis_label=fig_info["x_axis_label"],
                        y_axis_label=fig_info["y_axis_label"],
                    )
                )

            for user_index, user_id in enumerate(user_id_list):
                filtered_df = df[df["first_annotation_user_id"] == user_id]
                if filtered_df.empty:
                    logger.debug(f"dataframe is empty. user_id = {user_id}")
                    continue

                source = ColumnDataSource(data=filtered_df)
                color = self.my_palette[user_index]
                username = filtered_df.iloc[0]["first_annotation_username"]

                for fig, fig_info in zip(figs, fig_info_list):
                    self._plot_line_and_circle(
                        fig,
                        x_column_name=fig_info["x"],
                        y_column_name=fig_info["y"],
                        source=source,
                        legend_label=username,
                        color=color,
                    )

            hover_tool = self._create_hover_tool(tooltip_item)
            for fig in figs:
                self._set_legend(fig, hover_tool)

            bokeh.plotting.reset_output()
            bokeh.plotting.output_file(output_file, title=html_title)
            bokeh.plotting.save(bokeh.layouts.column(figs))

        tooltip_item = [
            "task_id",
            "phase",
            "status",
            "first_annotation_user_id",
            "first_annotation_username",
            "first_annotation_started_datetime",
            "updated_datetime",
            "annotation_worktime_hour",
            "inspection_worktime_hour",
            "acceptance_worktime_hour",
            "annotation_count",
            "input_data_count",
            "inspection_count",
        ]

        if len(df) == 0:
            logger.info("タスク一覧が0件のため出力しない")
            return

        user_id_list = self.create_user_id_list(
            df,
            "first_annotation_user_id",
            datetime_column="first_annotation_started_datetime",
            arg_user_id_list=first_annotation_user_id_list,
        )
        logger.debug(f"教師付者用の累積折れ線グラフに表示する、教師付者のuser_id = {user_id_list}")

        # 横軸が累計のアノテーション数
        fig_info_list_annotation_count = [
            dict(
                x="cumulative_annotation_count",
                y="annotation_worktime_hour/annotation_count",
                title="アノテーションあたり作業時間の変化",
                x_axis_label="アノテーション数",
                y_axis_label="教師付時間[hour]/アノテーション数",
            ),
            dict(
                x="cumulative_annotation_count",
                y="inspection_count/annotation_count",
                title="アノテーションあたり検査コメント数の変化",
                x_axis_label="アノテーション数",
                y_axis_label="検査コメント数/アノテーション数",
            ),
        ]
        write_cumulative_graph(fig_info_list_annotation_count, html_title="折れ線-横軸_アノテーション数-縦軸_アノテーションあたりの指標-教師付者用")

    def write_gradient_for_inspector(
        self,
        df: pandas.DataFrame,
        first_inspection_user_id_list: Optional[List[str]] = None,
    ):
        """
        日毎の生産性を検査作業者ごとにプロットする。

        Args:
            df:
            first_acceptance_user_id_list:

        Returns:

        """

        def write_graph(fig_info_list: List[Dict[str, str]], html_title: str):
            """

            Args:
                fig_info_list:
                html_title:

            Returns:

            """
            output_file = f"{self.line_graph_outdir}/{html_title}.html"
            logger.debug(f"{output_file} を出力します。")

            figs: List[bokeh.plotting.Figure] = []
            for fig_info in fig_info_list:
                figs.append(
                    figure(
                        plot_width=1200,
                        plot_height=600,
                        title=fig_info["title"],
                        x_axis_label=fig_info["x_axis_label"],
                        x_axis_type="datetime",
                        y_axis_label=fig_info["y_axis_label"],
                    )
                )

            for user_index, user_id in enumerate(user_id_list):
                filtered_df = df[df["first_inspection_user_id"] == user_id]
                if filtered_df.empty:
                    logger.debug(f"dataframe is empty. user_id = {user_id}")
                    continue

                source = ColumnDataSource(data=filtered_df)
                color = self.my_palette[user_index]
                username = filtered_df.iloc[0]["first_inspection_username"]

                for fig, fig_info in zip(figs, fig_info_list):
                    self._plot_line_and_circle(
                        fig,
                        x_column_name=fig_info["x"],
                        y_column_name=fig_info["y"],
                        source=source,
                        legend_label=username,
                        color=color,
                    )

            hover_tool = self._create_hover_tool(tooltip_item)
            for fig in figs:
                self._set_legend(fig, hover_tool)

            bokeh.plotting.reset_output()
            bokeh.plotting.output_file(output_file, title=html_title)
            bokeh.plotting.save(bokeh.layouts.column(figs))

        tooltip_item = [
            "task_id",
            "phase",
            "status",
            "first_inspection_user_id",
            "first_inspection_username",
            "first_inspection_started_datetime",
            "updated_datetime",
            "annotation_worktime_hour",
            "inspection_worktime_hour",
            "acceptance_worktime_hour",
            "annotation_count",
            "input_data_count",
        ]

        if len(df) == 0:
            logger.info("データが0件のため出力しない")
            return

        user_id_list = self.create_user_id_list(
            df,
            "first_inspection_user_id",
            datetime_column="first_inspection_started_datetime",
            arg_user_id_list=first_inspection_user_id_list,
        )
        logger.debug(f"検査者用の折れ線グラフに表示する、受入者のuser_id = {user_id_list}")

        fig_info_list_input_data_count = [
            dict(
                x="cumulative_annotation_count",
                y="annotation_worktime_hour/annotation_count",
                title="アノテーションあたり検査時間の変化",
                x_axis_label="アノテーション数",
                y_axis_label="検査時間[hour]/アノテーション数",
            ),
        ]
        write_graph(fig_info_list_input_data_count, html_title="折れ線-横軸_アノテーション数-縦軸_アノテーションあたりの指標-検査者用")

    def write_gradient_for_acceptor(
        self,
        df: pandas.DataFrame,
        first_acceptance_user_id_list: Optional[List[str]] = None,
    ):
        """
        日毎の生産性をプロットする

        Args:
            df:
            first_acceptance_user_id_list:

        Returns:

        """

        def write_graph(fig_info_list: List[Dict[str, str]], html_title: str):
            """

            Args:
                fig_info_list:
                html_title:

            Returns:

            """
            output_file = f"{self.line_graph_outdir}/{html_title}.html"
            logger.debug(f"{output_file} を出力します。")

            figs: List[bokeh.plotting.Figure] = []
            for fig_info in fig_info_list:
                figs.append(
                    figure(
                        plot_width=1200,
                        plot_height=600,
                        title=fig_info["title"],
                        x_axis_label=fig_info["x_axis_label"],
                        x_axis_type="datetime",
                        y_axis_label=fig_info["y_axis_label"],
                    )
                )

            for user_index, user_id in enumerate(user_id_list):
                filtered_df = df[df["first_acceptance_user_id"] == user_id]
                if filtered_df.empty:
                    logger.debug(f"dataframe is empty. user_id = {user_id}")
                    continue

                source = ColumnDataSource(data=filtered_df)
                color = self.my_palette[user_index]
                username = filtered_df.iloc[0]["first_acceptance_username"]

                for fig, fig_info in zip(figs, fig_info_list):
                    self._plot_line_and_circle(
                        fig,
                        x_column_name=fig_info["x"],
                        y_column_name=fig_info["y"],
                        source=source,
                        legend_label=username,
                        color=color,
                    )

            hover_tool = self._create_hover_tool(tooltip_item)
            for fig in figs:
                self._set_legend(fig, hover_tool)

            bokeh.plotting.reset_output()
            bokeh.plotting.output_file(output_file, title=html_title)
            bokeh.plotting.save(bokeh.layouts.column(figs))

        tooltip_item = [
            "task_id",
            "phase",
            "status",
            "first_acceptance_user_id",
            "first_acceptance_username",
            "first_acceptance_started_datetime",
            "updated_datetime",
            "annotation_worktime_hour",
            "inspection_worktime_hour",
            "acceptance_worktime_hour",
            "annotation_count",
            "input_data_count",
        ]

        if len(df) == 0:
            logger.info("データが0件のため出力しない")
            return

        user_id_list = self.create_user_id_list(
            df,
            "first_acceptance_user_id",
            datetime_column="first_acceptance_started_datetime",
            arg_user_id_list=first_acceptance_user_id_list,
        )
        logger.debug(f"受入者用の折れ線グラフに表示する、受入者のuser_id = {user_id_list}")

        fig_info_list_input_data_count = [
            dict(
                x="cumulative_annotation_count",
                y="acceptance_worktime_hour/annotation_count",
                title="アノテーションあたり受入時間の変化",
                x_axis_label="アノテーション数",
                y_axis_label="受入時間[hour]/アノテーション数",
            ),
        ]

        write_graph(fig_info_list_input_data_count, html_title="折れ線-横軸_アノテーション数-縦軸_アノテーションあたりの指標-受入者用")



"""
プロジェクトのユーザを表示する。
"""
import argparse
import logging
from typing import List

import requests
from annofabapi.models import ProjectMember

import annofabcli
from annofabcli import AnnofabApiFacade
from annofabcli.common.cli import AbstractCommandLineInterface, ArgumentParser, build_annofabapi_resource_and_login
from annofabcli.common.enums import FormatArgument
from annofabcli.common.visualize import AddProps

logger = logging.getLogger(__name__)


class ListUser(AbstractCommandLineInterface):
    """
    ユーザを表示する
    """

    def get_all_project_members(self, project_id: str, include_inactive: bool = False):
        query_params = {}
        if include_inactive:
            query_params.update({"include_inactive_member": ""})

        project_members = self.service.wrapper.get_all_project_members(project_id, query_params=query_params)
        return project_members

    def get_project_members_with_project_id(
        self, project_id_list: List[str], include_inactive: bool = False
    ) -> List[ProjectMember]:
        all_project_members: List[ProjectMember] = []

        for project_id in project_id_list:
            try:
                project, _ = self.service.api.get_project(project_id)
            except requests.exceptions.HTTPError as e:
                logger.warning(e)
                logger.warning(f"project_id = {project_id} のプロジェクトにアクセスできなかった（存在しないproject_id、またはプロジェクトメンバでない）")
                continue

            project_title = project["title"]
            project_members = self.get_all_project_members(project_id, include_inactive=include_inactive)
            logger.info(f"{project_title} のプロジェクトメンバを {len(project_members)} 件取得した。project_id={project_id}")

            for member in project_members:
                AddProps.add_properties_of_project(member, project_title)

            all_project_members.extend(project_members)

        return all_project_members

    def main(self):
        args = self.args

        project_id_list = annofabcli.common.cli.get_list_from_args(args.project_id)
        project_members = self.get_project_members_with_project_id(
            project_id_list,
            include_inactive=args.include_inactive,
        )

        self.print_according_to_format(project_members)


def main(args):
    service = build_annofabapi_resource_and_login(args)
    facade = AnnofabApiFacade(service)
    ListUser(service, facade, args).main()


def parse_args(parser: argparse.ArgumentParser):
    argument_parser = ArgumentParser(parser)

    list_group = parser.add_mutually_exclusive_group(required=True)
    list_group.add_argument(
        "-p",
        "--project_id",
        type=str,
        nargs="+",
        help="ユーザを表示するプロジェクトのproject_idを指定してください。`file://`を先頭に付けると、一覧が記載されたファイルを指定できます。",
    )

    parser.add_argument("--include_inactive", action="store_true", help="脱退されたメンバも表示します。")

    argument_parser.add_format(
        choices=[FormatArgument.CSV, FormatArgument.JSON, FormatArgument.PRETTY_JSON, FormatArgument.USER_ID_LIST],
        default=FormatArgument.CSV,
    )
    argument_parser.add_output()
    argument_parser.add_csv_format()
    argument_parser.add_query()

    parser.set_defaults(subcommand_func=main)


def add_parser(subparsers: argparse._SubParsersAction):
    subcommand_name = "list"
    subcommand_help = "複数のプロジェクトのプロジェクトメンバを表示する。"
    description = "複数のプロジェクトのプロジェクトメンバを表示する。"

    parser = annofabcli.common.cli.add_parser(subparsers, subcommand_name, subcommand_help, description)
    parse_args(parser)



import copy
import json
import logging.config
import os
import pkgutil
import re
import sys
from pathlib import Path
from typing import Any, Callable, Dict, List, Optional, Set, TypeVar

import dateutil.parser
import isodate
import pandas
import requests
import yaml

import annofabcli
from annofabcli.common.cli import DEFAULT_CSV_FORMAT
from annofabcli.common.enums import FormatArgument
from annofabcli.common.exceptions import AnnofabCliException

logger = logging.getLogger(__name__)

T = TypeVar("T")  # Can be anything


def read_lines(filepath: str) -> List[str]:
    """ファイルを行単位で読み込む。改行コードを除く"""
    with open(filepath) as f:
        lines = f.readlines()
    return [e.rstrip("\r\n") for e in lines]


def read_lines_except_blank_line(filepath: str) -> List[str]:
    """ファイルを行単位で読み込む。ただし、改行コード、空行を除く"""
    lines = read_lines(filepath)
    return [line for line in lines if line != ""]


def _is_file_scheme(value: str):
    return value.startswith("file://")


def load_logging_config(log_dir: str, logging_yaml_file: Optional[str] = None):
    """
    ログ設定ファイルを読み込み、loggingを設定する。

    Args:
        log_dir: ログの出力先
        logging_yaml_file: ログ設定ファイル。指定しない場合、`data/logging.yaml`を読み込む. 指定した場合、log_dir, log_filenameは無視する。

    """

    if logging_yaml_file is not None:
        if os.path.exists(logging_yaml_file):
            with open(logging_yaml_file, encoding="utf-8") as f:
                logging_config = yaml.safe_load(f)
                logging.config.dictConfig(logging_config)
        else:
            logger.warning(f"{logging_yaml_file} does not exist.")
            set_default_logger(log_dir)

    else:
        set_default_logger(log_dir)


def set_default_logger(log_dir: str = ".log", log_filename: str = "annofabcli.log"):
    """
    デフォルトのロガーを設定する。パッケージ内のlogging.yamlを読み込む。
    """
    data = pkgutil.get_data("annofabcli", "data/logging.yaml")
    if data is None:
        logger.warning("annofabcli/data/logging.yaml が読み込めませんでした")
        raise AnnofabCliException("annofabcli/data/logging.yaml が読み込めませんでした")

    logging_config = yaml.safe_load(data.decode("utf-8"))

    log_filename = f"{str(log_dir)}/{log_filename}"
    logging_config["handlers"]["fileRotatingHandler"]["filename"] = log_filename
    Path(log_dir).mkdir(exist_ok=True, parents=True)

    logging.config.dictConfig(logging_config)


def duplicated_set(target_list: List[T]) -> Set[T]:
    """
    重複しているsetを返す
    Args:
        target_list: 確認するList

    Returns:
        重複しているset

    """
    return {x for x in set(target_list) if target_list.count(x) > 1}


def progress_msg(index: int, size: int):
    """
    `1/100件目`という進捗率を表示する
    """
    digit = len(str(size))
    str_format = f"{{:{digit}}} / {{:{digit}}} 件目"
    return str_format.format(index, size)


def output_string(target: str, output: Optional[str] = None) -> None:
    """
    文字列を出力する。

    Args:
        target: 出力対象の文字列
        output: 出力先。Noneなら標準出力に出力する。
    """
    if output is None:
        print(target)
    else:
        Path(output).parent.mkdir(parents=True, exist_ok=True)
        with open(output, mode="w", encoding="utf_8") as f:
            f.write(target)
            logger.info(f"{output} に出力しました。")


def print_json(target: Any, is_pretty: bool = False, output: Optional[str] = None) -> None:
    """
    JSONを出力する。

    Args:
        target: 出力対象のJSON
        is_pretty: 人が見やすいJSONを出力するか
        output: 出力先。Noneなら標準出力に出力する。

    """
    if is_pretty:
        output_string(json.dumps(target, indent=2, ensure_ascii=False), output)
    else:
        output_string(json.dumps(target, ensure_ascii=False), output)


def print_csv(df: pandas.DataFrame, output: Optional[str] = None, to_csv_kwargs: Optional[Dict[str, Any]] = None):
    if output is not None:
        Path(output).parent.mkdir(parents=True, exist_ok=True)

    path_or_buf = sys.stdout if output is None else output

    kwargs = copy.deepcopy(DEFAULT_CSV_FORMAT)
    if to_csv_kwargs is None:
        df.to_csv(path_or_buf, **kwargs)
    else:
        kwargs.update(to_csv_kwargs)
        df.to_csv(path_or_buf, **kwargs)

    if output is not None:
        logger.info(f"{output} に出力しました。")


def print_id_list(id_list: List[Any], output: Optional[str]):
    s = "\n".join(id_list)
    output_string(s, output)


def print_according_to_format(
    target: Any, arg_format: FormatArgument, output: Optional[str] = None, csv_format: Optional[Dict[str, Any]] = None
):
    """
    コマンドライン引数 ``--format`` の値にしたがって、内容を出力する。

    Args:
        target: 出力する内容
        format: 出力フォーマット
        output: 出力先（オプション）
        csv_format: CSVのフォーマット（オプション）


    """

    if arg_format == FormatArgument.PRETTY_JSON:
        annofabcli.utils.print_json(target, is_pretty=True, output=output)

    elif arg_format == FormatArgument.JSON:
        annofabcli.utils.print_json(target, is_pretty=False, output=output)

    elif arg_format == FormatArgument.CSV:
        df = pandas.DataFrame(target)
        annofabcli.utils.print_csv(df, output=output, to_csv_kwargs=csv_format)

    elif arg_format == FormatArgument.TASK_ID_LIST:
        task_id_list = [e["task_id"] for e in target]
        print_id_list(task_id_list, output)

    elif arg_format == FormatArgument.PROJECT_ID_LIST:
        project_id_list = [e["project_id"] for e in target]
        print_id_list(project_id_list, output)

    elif arg_format == FormatArgument.INPUT_DATA_ID_LIST:
        input_data_id_list = [e["input_data_id"] for e in target]
        print_id_list(input_data_id_list, output)

    elif arg_format == FormatArgument.USER_ID_LIST:
        user_id_list = [e["user_id"] for e in target]
        print_id_list(user_id_list, output)

    elif arg_format == FormatArgument.INSPECTION_ID_LIST:
        inspection_id_list = [e["inspection_id"] for e in target]
        print_id_list(inspection_id_list, output)


def to_filename(s: str):
    """
    文字列をファイル名に使えるよう変換する。ファイル名に使えない文字は"__"に変換する。
    Args:
        s:

    Returns:
        ファイル名用の文字列

    """
    return re.sub(r'[\\|/|:|?|.|"|<|>|\|]', "__", s)


def is_file_scheme(str_value: str) -> bool:
    """
    file schemaかどうか

    """
    return str_value.startswith("file://")


def get_file_scheme_path(str_value: str) -> Optional[str]:
    """
    file schemaのパスを取得する。file schemeでない場合は、Noneを返す

    """
    if is_file_scheme(str_value):
        return str_value[len("file://") :]
    else:
        return None


def isoduration_to_hour(duration) -> float:
    """
    ISO 8601 duration を 時間に変換する
    Args:
        duration (str): ISO 8601 Durationの文字

    Returns:
        変換後の時間。

    """
    return isodate.parse_duration(duration).total_seconds() / 3600


def isoduration_to_minute(duration) -> float:
    """
    ISO 8601 duration を 分に変換する

    Args:
        duration (str): ISO 8601 Durationの文字
    Returns:
        変換後の分
    """
    return isodate.parse_duration(duration).total_seconds() / 60


def datetime_to_date(str_datetime: str) -> str:
    """
    ISO8601形式の日時を日付に変換する

    Args:
        str_datetime: ISO8601の拡張形式（YYYY-MM-DDThh:mm:ss+09:00）

    Returns:
        日時(YYYY-MM-DD)
    """
    return str(dateutil.parser.parse(str_datetime).date())


def allow_404_error(function):
    """
    Not Found Error(404)を無視(許容)して、処理する。Not Foundのとき戻りはNoneになる。
    リソースの存在確認などに利用する。
    try-exceptを行う。また404 Errorが発生したときのエラーログを無効化する
    """

    def wrapped(*args, **kwargs):
        annofabapi_logger_level = logging.getLogger("annofabapi").level
        backoff_logger_level = logging.getLogger("backoff").level

        try:
            # 不要なログが出力されないようにする
            logging.getLogger("annofabapi").setLevel(level=logging.INFO)
            logging.getLogger("backoff").setLevel(level=logging.CRITICAL)

            return function(*args, **kwargs)

        except requests.exceptions.HTTPError as e:
            if e.response.status_code != requests.codes.not_found:
                raise e
        finally:
            # ロガーの設定を元に戻す
            logging.getLogger("annofabapi").setLevel(level=annofabapi_logger_level)
            logging.getLogger("backoff").setLevel(level=backoff_logger_level)

    return wrapped


def get_cache_dir() -> Path:
    """
    環境変数から、annofabcliのキャシュディレクトリを取得する。
    キャッシュホームディレクトリは環境変数 ``XDG_CACHE_HOME`` から取得する。デフォルトは ``$HOME/.cache``である。

    Returns:
        annofabcliのキャッシュディレクトリ

    """
    cache_home_dir = os.environ.get("XDG_CACHE_HOME")
    if cache_home_dir is None:
        cache_home_dir_path = Path.home() / ".cache"
    else:
        cache_home_dir_path = Path(cache_home_dir)

    return cache_home_dir_path / "annofabcli"


def read_multiheader_csv(csv_file: str, header_row_count: int = 2, **kwargs) -> pandas.DataFrame:
    """
    複数ヘッダ行のCSVを読み込む。その際、"Unnnamed"の列名は空文字に変更する。

    Args:
        csv_file:
        header_row_count: ヘッダの行数。2以上を指定する。

    Returns:
        pandas.DataFrame

    """
    kwargs["header"] = list(range(header_row_count))
    df = pandas.read_csv(csv_file, **kwargs)
    for level in range(0, header_row_count):
        columns = df.columns.levels[level]
        rename_columns = {c: "" for c in columns if re.fullmatch(r"Unnamed: .*", c) is not None}
        df.rename(
            columns=rename_columns,
            level=level,
            inplace=True,
        )
    return df


def _catch_exception(function: Callable[..., Any]) -> Callable[..., Any]:
    """
    Exceptionをキャッチしてログにstacktraceを出力する。
    """

    def wrapped(*args, **kwargs):
        try:
            return function(*args, **kwargs)
        except Exception as e:  # pylint: disable=broad-except
            logger.warning(e)
            logger.exception(e)

    return wrapped
